import datetime
import importlib
import os
import re
import subprocess
import sys
import threading
import unicodedata
from copy import deepcopy
from shlex import split
from site import getsitepackages
from sys import platform
from time import sleep

import jsonpickle
import numpy
import pandas
import xarray as xr

from pyplan_core import cubepy
from pyplan_core.classes.BaseNode import BaseNode
from pyplan_core.classes.evaluators.Evaluator import Evaluator
from pyplan_core.classes.Intellisense import Intellisense
from pyplan_core.classes.IOModule import IOModule
from pyplan_core.classes.PyplanFunctions import PyplanFunctions, Selector
from pyplan_core.classes.wizards import (CalculatedField, DataarrayFromPandas,
                                         DataframeGroupby, DataframeIndex,
                                         InputTable, SelectColumns, SelectRows,
                                         sourcecsv, CreateIndex, IndexFromPandas, DataArrayFilter)

from .DefaultNodeFormats import default_formats


class Model(object):

    DEFAULT_IMPORTS = {
        'np': numpy,
        'pd': pandas,
        'cubepy': cubepy,
        'xr': xr
    }

    def __init__(self, WSClass=None):
        self._nodeDic = {}
        self._modelProp = {}
        self._modelNode = None
        self._isLoadingModel = False
        self.initialize()
        self.evaluationVersion = 0
        self.inCyclicEvaluate = False
        self._scenarioDic = dict()
        self._wizard = None
        self._currentProcessingNode = ''
        self._currentInstallProgress = []
        self._customImports = None
        self.company_code = None
        self.session_key = None
        self.ws = None
        self.debugMode = None
        self.WS = WSClass

    # Props

    @property
    def nodeDic(self):
        return self._nodeDic

    @property
    def modelProp(self):
        return self._modelProp

    @property
    def modelNode(self):
        return self._modelNode

    @property
    def isLoadingModel(self):
        return self._isLoadingModel

    # Methods

    def getPID(self):
        return os.getpid()

    def getDefaultNodeFormat(self, nodeClass):
        default_formats
        if nodeClass in default_formats:
            return deepcopy(default_formats[nodeClass])
        else:
            return None

    def getTotalMemory(self):
        res = 0
        for node in self.nodeDic:
            res = res+self.nodeDic[node].usedMemory
        return res

    def getCurrentModelPath(self):
        if self.existNode('current_path'):
            return self.getNode('current_path').result
        return ''

    def setCurrentModelPath(self, value):
        if self.existNode('current_path'):
            self.getNode('current_path').definition = 'result="""' + \
                str(value) + '"""'

    def currentProcessingNode(self, nodeId):
        if nodeId not in ['__evalnode__', 'current_path']:
            self._currentProcessingNode = nodeId

    def initialize(self, modelName=None):

        if modelName is None:
            self._modelNode = self.createNode('new_model', 'model', '_model_')
        else:
            newId = modelName.lower()
            newId = re.sub('[^0-9a-z]+', '_', newId)
            self._modelNode = self.createNode(newId, 'model', '_model_')
            self._modelNode.title = modelName

        self._scenarioDic = dict()
        self._wizard = None

    def connectToWS(self, company_code, session_key):
        # Connect to WS
        self.company_code = company_code
        self.session_key = session_key
        if self.WS:
            self.ws = self.WS(company_code=company_code,
                              session_key=session_key)

    def createNode(self, identifier=None, nodeClass=None, moduleId=None, x=None, y=None, toObj=False, originalId=None):
        """Create new node"""
        newNode = BaseNode(model=self, identifier=identifier, nodeClass=nodeClass,
                           moduleId=moduleId, x=x, y=y, originalId=originalId)
        id = newNode.identifier
        self.nodeDic[id] = newNode
        newNode = None
        if toObj:
            return self.nodeDic[id].toObj()
        else:  # uso interno
            return self.nodeDic[id]

    def deleteNodes(self, nodes, removeAliasIfNotIn=None):
        """Delete nodes by node id"""
        if not nodes is None:
            for nodeId in nodes:
                if self.existNode(nodeId) and nodeId != '_model_':
                    # check for module
                    if self.getNode(nodeId).nodeClass == 'module':
                        childs = self.findNodes('moduleId', nodeId)
                        childsIds = [c.identifier for c in childs]
                        self.deleteNodes(childsIds, removeAliasIfNotIn)

                    # check for aliases
                    aliases = []
                    aliases = self.findNodes('originalId', nodeId)
                    if aliases:
                        aliasesId = [a.identifier for a in aliases]
                        # check for aliases in other modules
                        if removeAliasIfNotIn:
                            _auxAliases = []
                            for _aliasId in aliasesId:
                                if self.isIn(_aliasId, removeAliasIfNotIn):
                                    _auxAliases.append(_aliasId)
                            if len(_auxAliases) > 0:
                                self.deleteNodes(
                                    _auxAliases, removeAliasIfNotIn)
                        else:
                            self.deleteNodes(aliasesId, removeAliasIfNotIn)

                    self.nodeDic[nodeId].preDelete()
                    self.nodeDic[nodeId].release()
                    self.nodeDic[nodeId] = None
                    del self.nodeDic[nodeId]
        pass

    def existNode(self, nodeId):
        """Return True if node exists in model"""
        nodeId = self.clearId(nodeId)
        if (not nodeId is None) & (nodeId in self.nodeDic):
            return True
        return False

    def getNode(self, nodeId):
        """Renor node from node dictionary"""
        if self.existNode(nodeId):
            return self.nodeDic[nodeId]

    def isChild(self, nodeId, modulesId):
        """Return true if nodeid or one of your parents is in any of modulesId modules"""
        res = False
        if self.existNode(nodeId):
            aux = self.getNode(nodeId).moduleId
            nChance = 20
            while (res == False and aux != '_model_' and nChance > 0):
                res = aux in modulesId
                node = self.getNode(aux)
                if node:
                    aux = node.moduleId
                else:
                    nChance = 0
                nChance -= 1
        return res

    def getAPlace(self, moduleId, x, y, w, h, deltaX=110):
        """ Get a place for insert a node """
        res = dict(x=x, y=y)
        options = 100
        l1x = int(x)
        l1y = int(y)
        r1x = int(x) + int(w)
        r1y = int(y) + int(h)
        endy = int(y) + int(h)
        while options > 0:
            options -= 1
            found = False
            for node in self.findNodes('moduleId', moduleId):
                if node.nodeClass != "text":
                    l2x = int(node.x)
                    l2y = int(node.y)
                    r2x = int(node.x) + int(node.w)
                    r2y = int(node.y) + int(node.h)
                    if not (l1x > r2x or l2x > r1x or l1y > r2y or l2y > r1y):
                        found = True
                        break
            if found:
                l1x = l1x + deltaX
                r1x = r1x + deltaX
            else:
                res["x"] = l1x
                res["y"] = l1y
                break
        return res

    def isNodeInScenario(self, nodeId):
        if nodeId in self._scenarioDic:
            return True
        else:
            return False

    def evaluateNode(self, nodeId, dims=None, rows=None, columns=None, summaryBy='sum', bottomTotal=False, rightTotal=False, fromRow=0, toRow=0, resultType=""):
        """Evaluate node. Call evaluator class for implement diferent evaluators."""
        if self.existNode(nodeId):
            result = None
            if nodeId in self._scenarioDic:
                result = self._scenarioDic[nodeId]
            else:
                result = self.nodeDic[nodeId].result

            if not result is None:
                self.evaluationVersion += 1
                evaluator = Evaluator.createInstance(result)
                if not evaluator.checkStructure(result, resultType):
                    raise ValueError("bad_node_structure")
                    #raise exceptions.NotAcceptable("bad_node_structure")
                return evaluator.evaluateNode(result, self.nodeDic, nodeId, dims, rows, columns, summaryBy, bottomTotal, rightTotal, fromRow, toRow)
        return ''

    def executeButton(self, nodeId):
        """Execute node of class button"""
        if self.existNode(nodeId):
            self.nodeDic[nodeId].invalidate()
            toReturn = self.nodeDic[nodeId].result
            if toReturn is None:
                toReturn = ''
            return toReturn

        return ''

    def previewNode(self, nodeId, debugMode=""):
        """Perform preview of a node"""
        try:

            result = None
            if self.existNode(nodeId):
                if debugMode:
                    self.debugMode = debugMode
                    if debugMode == "node":
                        self.nodeDic[nodeId].silentInvalidate()
                    elif debugMode == "model":
                        for node_key in self.nodeDic:
                            self.nodeDic[node_key].silentInvalidate()

                if not self.nodeDic[nodeId].originalId is None:
                    nodeId = self.nodeDic[nodeId].originalId
                if self.nodeDic[nodeId].nodeClass in ["button", "module", "text"]:
                    self.evaluationVersion += 1
                    _dummy = self.nodeDic[nodeId].result  # for execute button
                elif not self.nodeDic[nodeId].result is None:
                    self.evaluationVersion += 1
                    evaluator = Evaluator.createInstance(
                        self.nodeDic[nodeId].result)
                    result = evaluator.previewNode(self.nodeDic, nodeId)
            if result is None:
                evaluator = Evaluator.createInstance(None)
                result = evaluator.generateEmptyPreviewResponse(
                    self.nodeDic, nodeId)
            return result
        finally:
            if self.debugMode and self.ws:
                self.ws.sendDebugInfo("endPreview", "", "endPreview")
            self.debugMode = None

    def getCubeValues(self, query):
        """Evaluate node. Used for pivotgrid"""
        nodeId = query['cube']
        if self.existNode(nodeId):

            result = None
            if nodeId in self._scenarioDic:
                result = self._scenarioDic[nodeId]
            else:
                result = self.nodeDic[nodeId].result

            if not result is None:
                evaluator = Evaluator.createInstance(result)
                return evaluator.getCubeValues(result, self.nodeDic, nodeId, query)

    def getCubeDimensionValues(self, query):
        """Return the values of a dimension of node. Used from pivotgrid"""
        nodeId = query['cube']
        if self.existNode(nodeId):
            result = None
            if nodeId in self._scenarioDic:
                result = self._scenarioDic[nodeId]
            else:
                result = self.nodeDic[nodeId].result

            if not result is None:
                evaluator = Evaluator.createInstance(result)
                return evaluator.getCubeDimensionValues(result, self.nodeDic, nodeId, query)

    def getCubeMetadata(self, nodeId, resultType=''):
        """Return metadata of cube. Used from pivotgrid"""
        if self.existNode(nodeId):
            result = None
            if nodeId in self._scenarioDic:
                result = self._scenarioDic[nodeId]
            else:
                result = self.nodeDic[nodeId].result

            if not result is None:
                evaluator = Evaluator.createInstance(result)
                if resultType:
                    if not evaluator.checkStructure(result, resultType):
                        raise ValueError("bad_node_structure")
                return evaluator.getCubeMetadata(result, self.nodeDic, nodeId)

    def setNodeValueChanges(self, changes):
        """Set values for node using filters"""
        nodeId = changes['node']
        if self.existNode(nodeId):
            if self.nodeDic[nodeId].nodeClass == 'formnode':
                nodeId = self.nodeDic[nodeId].originalId
                evaluator = Evaluator.createInstance(None)
                return evaluator.setNodeValueChanges(self.nodeDic, nodeId, changes)
            else:
                if not self.nodeDic[nodeId].originalId is None:
                    nodeId = self.nodeDic[nodeId].originalId
                if not self.nodeDic[nodeId].result is None:
                    evaluator = Evaluator.createInstance(
                        self.nodeDic[nodeId].result)
                    return evaluator.setNodeValueChanges(self.nodeDic, nodeId, changes)

    def getDiagram(self, moduleId=None):
        """Get diagram"""
        if moduleId is None:
            moduleId = self.modelNode.identifier
        moduleId = self.clearId(moduleId)
        res = {
            'moduleId': moduleId,
            'arrows': [],
            'nodes': [],
            'breadcrumb': self.getBreadcrumb(moduleId)
        }
        nodeList = self.findNodes('moduleId', moduleId)
        nodeList.sort(key=lambda x: int(x.z))
        for node in nodeList:
            res['nodes'].append(node.toObj(
                exceptions=['definition'], fillDefaultProperties=True))
        return res

    def getBreadcrumb(self, moduleId=None):
        """Get breadcrumb"""
        if moduleId is None:
            moduleId = self.modelNode.identifier
        moduleId = self.clearId(moduleId)
        res = []
        aux = moduleId
        while aux != self.modelNode.identifier and self.existNode(aux):
            res.append({
                'identifier': aux,
                'title': (self.getNode(aux).title or aux)
            })
            aux = self.getNode(aux).moduleId

        res.append({'identifier': self.modelNode.identifier,
                    'title': self.modelNode.title or 'Main'})
        return res

    def isIn(self, nodeId, moduleId):
        """ Return true if nodeId is in moduleId. Search for parents"""
        res = False
        aux = nodeId
        _secure = 1
        while aux != self.modelNode.identifier and self.existNode(aux) and _secure < 100:
            if aux == moduleId:
                res = True
                break
            aux = self.getNode(aux).moduleId
            _secure += 1
        return res

    def isSelector(self, nodeId):
        """Return True if nodeId is of type Selector"""
        if self.existNode(nodeId):
            return isinstance(self.getNode(nodeId).result, Selector)
        return False

    def getSelector(self, nodeId):
        """Return selector data if node is of type selector"""
        if self.isSelector(nodeId):
            return self.getNode(nodeId).result.toObj()

    def setSelectorValue(self, nodeId, value):
        """Set selector value if node is of type selector"""
        if self.isSelector(nodeId):

            node = self.getNode(nodeId)
            selector = node.result

            if not selector.isSameValue(value):
                definition = node.definition
                new_definition = selector.generateDefinition(definition, value)
                if new_definition:
                    node.definition = new_definition

    def release(self):
        """Release model. Free all resources """
        if not self._modelNode is None:
            self._modelNode.release()

        (xx.release() for xx in self.nodeDic)
        keys = [x for x in self.nodeDic]
        for key in keys:
            del self.nodeDic[key]

        self._nodeDic = {}
        self._modelProp = {}
        self._modelNode = None
        self._scenarioDic = dict()
        self._wizard = None
        self._customImports = None

        return

    def getNextIdentifier(self, prefix):
        """Get next free identifier of node"""
        reg = r'(\d+$)'
        matches = re.findall(reg, prefix)
        start_at = 1
        if len(matches) > 0:
            num = matches[0]
            start_at = int(num)+1
            if start_at > 100000000:
                prefix += '_'
                start_at = 1
            else:
                prefix = prefix[: -len(num)]

        for num in range(start_at, 100000000):
            key = prefix+str(num)
            if not key in self.nodeDic:
                return key

    def clearId(self, nodeId):
        """DEPRECATED"""
        # if not nodeId is None:
        #     return nodeId.lower()
        return nodeId

    def updateNodeIdInDic(self, oldNodeId, newNodeId):
        """Update node identifier on all dictionary."""
        if self.existNode(oldNodeId):
            newNodeId = self.clearId(newNodeId)
            self.nodeDic[newNodeId] = self.nodeDic[oldNodeId]
            for node in self.findNodes('moduleId', oldNodeId):
                node.moduleId = newNodeId
            for node in self.findNodes('originalId', oldNodeId):
                node.originalId = newNodeId
                node._definition = 'result = ' + str(newNodeId)

            del self.nodeDic[oldNodeId]
            return True
        else:
            return False

    def setNodeProperties(self, nodeId, properties):
        """Update properties of a node"""
        nodeId = self.clearId(nodeId)
        if self.existNode(nodeId):
            _node = self.getNode(nodeId)
            for prop in properties:
                if '.' in prop['name']:
                    nodeProp, objProp = prop['name'].split('.')
                    setattr(getattr(_node, nodeProp), objProp, prop['value'])
                else:
                    setattr(_node, prop['name'], prop['value'])

    def getNodeProperties(self, nodeProperties):
        """Get properties of a node"""
        if (not nodeProperties is None) and (nodeProperties['node'] != ''):
            nodeId = self.clearId(nodeProperties['node'])
            if self.existNode(nodeId):
                _node = self.getNode(nodeId)
                for prop in nodeProperties['properties']:
                    if hasattr(_node, prop['name']):
                        prop['value'] = getattr(_node, prop['name'])
                return nodeProperties
        pass

    def setModelProperties(self, properties):
        """Update properties of model"""
        for key in properties:
            if key == 'modelId':
                continue
            elif key in ['identifier', 'title']:
                setattr(self.modelNode, key, properties[key])
            else:
                self.modelProp[key] = properties[key]

    def getModelProperties(self):
        """Get model propierties"""
        res = dict()
        # fill model id and tile
        res['identifier'] = self.modelNode.identifier
        res['title'] = self.modelNode.title

        for key in self.modelProp:
            res[key] = self.modelProp[key]

        return res

    def getIndexes(self, nodeId):
        """Return indexes of a node"""
        if self.existNode(nodeId):
            if nodeId in self._scenarioDic:
                evaluator = Evaluator.createInstance(self._scenarioDic[nodeId])
                return evaluator.getIndexes(self.nodeDic[nodeId], self._scenarioDic[nodeId])
            else:
                _node = self.getNode(nodeId)
                return _node.indexes

    def getIndexValues(self, data):
        """Return values of a index node."""
        tmpNodeId = data.index_id if data.node_id is None or data.node_id == '' else data.node_id
        if self.existNode(tmpNodeId):
            result = self.nodeDic[tmpNodeId].result
            if data.node_id in self._scenarioDic:
                result = self._scenarioDic[data.node_id]
            evaluator = Evaluator.createInstance(result)
            return evaluator.getIndexValues(self.nodeDic, data, result)

    def getIndexType(self, nodeId, indexId):
        """Return index type"""
        tmpNodeId = indexId if nodeId is None or nodeId == '' else nodeId
        if self.existNode(tmpNodeId):
            evaluator = Evaluator.createInstance(
                self.nodeDic[tmpNodeId].result)
            return evaluator.getIndexType(self.nodeDic, nodeId, indexId)

    def getIndexesWithLevels(self, nodeId):
        """Return indexes of a node"""
        if self.existNode(nodeId):
            if nodeId in self._scenarioDic:
                evaluator = Evaluator.createInstance(self._scenarioDic[nodeId])
                return evaluator.getIndexesWithLevels(self.nodeDic[nodeId], self._scenarioDic[nodeId])
            else:
                evaluator = Evaluator.createInstance(
                    self.nodeDic[nodeId].result)
                return evaluator.getIndexesWithLevels(self.nodeDic[nodeId])

    def isTable(self, nodeId):
        """return true if node is a table"""
        res = '0'
        if self.existNode(nodeId):
            evaluator = Evaluator.createInstance(self.nodeDic[nodeId].result)
            res = evaluator.isTable(self.getNode(nodeId))
        return res

    def getArrows(self, moduleId):
        """Return all arrows of moduleId"""
        res = []
        modulesInLevel = []
        inputsInOtherLevel = []
        outputsInOtherLevel = []
        thisLevel = self.findNodes('moduleId', moduleId)
        thisIds = [node.identifier for node in thisLevel]
        for node in thisLevel:
            if node.nodeClass == 'module':
                modulesInLevel.append(node.identifier)

        # node to node
        for node in thisLevel:
            if node.nodeClass not in ['module', 'text']:
                for outputNodeId in node.outputs:
                    # aliases
                    fullOutputs = []
                    fullOutputs = self.findNodes('originalId', outputNodeId)
                    fullOutputs.append(self.getNode(outputNodeId))
                    for o in fullOutputs:
                        if not o is None:
                            element = {'from': node.identifier,
                                       'to': o.identifier}
                            if o.identifier in thisIds:
                                if node.nodeInfo.showOutputs and o.nodeInfo.showInputs:
                                    if self.existArrow(element['from'], element['to'], res) == False:
                                        res.append(element)
                            elif o.identifier not in thisIds:
                                if self.existArrow(element['from'], element['to'], outputsInOtherLevel) == False:
                                    # if theres an alias in this level don't include the arrow
                                    if not len(self.getAliasInLevel(o.identifier, moduleId)) > 0:
                                        outputsInOtherLevel.append(element)
                for inputNodeId in node.inputs:
                    # aliases
                    fullInputs = []
                    fullInputs = self.findNodes('originalId', inputNodeId)
                    fullInputs.append(self.getNode(inputNodeId))
                    for i in fullInputs:
                        if not i is None:
                            element = {'from': i.identifier,
                                       'to': node.identifier}
                            if i.identifier in thisIds:
                                if i.nodeInfo.showOutputs and node.nodeInfo.showInputs:
                                    if self.existArrow(element['from'], element['to'], res) == False:
                                        res.append(element)
                            elif i.identifier not in thisIds:
                                if self.existArrow(element['from'], element['to'], inputsInOtherLevel) == False:
                                    # if theres an alias in this level don't include the arrow
                                    if not len(self.getAliasInLevel(i.identifier, moduleId)) > 0:
                                        inputsInOtherLevel.append(element)

        # node to module
        if outputsInOtherLevel:
            for d in outputsInOtherLevel:
                newTo = []
                nodeFrom = d['from']
                nodeTo = d['to']
                if self.getNode(nodeTo).isin in self.nodeDic:
                    newTo = self.getParentModule(nodeTo, modulesInLevel)
                if newTo:
                    element = {'from': nodeFrom, 'to': newTo}
                    if self.getNode(nodeFrom).nodeInfo.showOutputs and self.getNode(newTo).nodeInfo.showInputs:
                        if self.existArrow(element['from'], element['to'], res) == False:
                            res.append(element)

        # module to node
        if inputsInOtherLevel:
            for d in inputsInOtherLevel:
                newFrom = []
                nodeFrom = d['from']
                nodeTo = d['to']
                if self.getNode(nodeFrom).isin in self.nodeDic:
                    newFrom = self.getParentModule(nodeFrom, modulesInLevel)
                if newFrom:
                    element = {'from': newFrom, 'to': nodeTo}
                    if self.getNode(newFrom).nodeInfo.showOutputs and self.getNode(nodeTo).nodeInfo.showInputs:
                        if self.existArrow(element['from'], element['to'], res) == False:
                            res.append(element)

        # module to module
        modulesComplete = []
        for mod in modulesInLevel:
            modulesComplete.append(
                {'module': mod, 'nodes': self.getNodesInModule(mod, [])})

        for mod in modulesComplete:
            for node in mod['nodes']:
                if self.getNode(mod['module']).nodeInfo.showOutputs:
                    tempOutputs = node.outputs
                    if tempOutputs:
                        for output in tempOutputs:
                            for auxModule in modulesComplete:
                                if(mod['module'] != auxModule['module'] and self.getNode(auxModule['module']).nodeInfo.showInputs):
                                    # module to module
                                    if self.getNode(output) in auxModule['nodes']:
                                        element = {
                                            'from': mod['module'], 'to': auxModule['module']}
                                        if self.existArrow(element['from'], element['to'], res) == False:
                                            res.append(element)
                                """# module to alias
                                aliases = self.getAliasInLevel(
                                    output, moduleId)
                                if len(aliases) > 0:
                                    for alias in aliases:
                                        if alias.nodeInfo.showInputs:
                                            element = {
                                                "from": mod["module"], "to": alias.identifier}
                                            if self.existArrow(element["from"], element["to"], res) == False:
                                                res.append(element)
                                            # alias to alias
                                            outputAliases = self.getAliasInLevel(
                                                node.identifier, moduleId)
                                            if len(outputAliases) > 0:
                                                for outAl in outputAliases:
                                                    if outAl.nodeInfo.showOutputs:
                                                        element = {
                                                            "from": outAl.identifier, "to": alias.identifier}
                                                        if self.existArrow(element["from"], element["to"], res) == False:
                                                            res.append(element)

                # alias to module
                if self.getNode(mod["module"]).nodeInfo.showInputs:
                    tempInputs = node.inputs
                    if tempInputs:
                        for inp in tempInputs:
                            aliases = self.getAliasInLevel(inp, moduleId)
                            if len(aliases) > 0:
                                for auxModule in modulesComplete:
                                    for alias in aliases:
                                        if alias.nodeInfo.showOutputs:
                                            element = {
                                                "from": alias.identifier, "to": mod["module"]}
                                            if self.existArrow(element["from"], element["to"], res) == False:
                                                res.append(element)"""

        return res

    def existArrow(self, aFrom, aTo, arrowsList):
        """Return true if exists de arrow from-to"""
        if arrowsList:
            for arrow in arrowsList:
                if arrow['from'] == aFrom and arrow['to'] == aTo:
                    return True
        return False

    def getAliasInLevel(self, nodeIdentifier, levelId):
        """Returns the aliases in the level"""
        res = []
        aliasNodes = self.findNodes('originalId', nodeIdentifier)
        if aliasNodes is not None:
            for alias in aliasNodes:
                if alias.moduleId == levelId:
                    res.append(alias)
                    break
        return res

    def getParentModulesWithAlias(self, moduleId, modulesArray):
        """Return the parent module with alias"""
        if moduleId != '_model_':
            if moduleId not in modulesArray:
                modulesArray.append(moduleId)
            alias = []
            alias = self.findNodes('originalId', moduleId)
            if alias:
                for a in alias:
                    if a.identifier not in modulesArray:
                        modulesArray.append(a.identifier)
            if self.getNode(moduleId).isin in self.nodeDic:
                return self.getParentModulesWithAlias(self.getNode(moduleId).isin, modulesArray)
            else:
                return modulesArray
        else:
            return modulesArray

    def getParentModule(self, moduleId, modulesInLevel):
        """Return parent module"""
        if moduleId == '_model_':
            return None
        else:
            if moduleId in modulesInLevel:
                return moduleId
            else:
                if self.getNode(moduleId).isin in self.nodeDic:
                    return self.getParentModule(self.getNode(moduleId).isin, modulesInLevel)
                else:
                    return None

    def getNodesInModule(self, moduleId, nodesInSubLevels):
        """Return nodes un module"""

        subLevelNodes = []
        modulesInSubLevels = []
        subLevelNodes = self.findNodes('moduleId', moduleId)
        if subLevelNodes:
            for node in subLevelNodes:
                if node.nodeClass == 'module':
                    modulesInSubLevels.append(node)
                else:
                    if node.nodeClass != 'text':
                        nodesInSubLevels.append(node)

            if modulesInSubLevels:
                for module in modulesInSubLevels:
                    self.getNodesInModule(module.identifier, nodesInSubLevels)

        return nodesInSubLevels

    # UNUSED FULL ARROWS (NODE TO NODE WITH ALIASES)
    """
    def getArrows(self,moduleId):
        res=[]
        thisLevel = self.findNodes("moduleId",moduleId)
        thisIds = [node.identifier for node in thisLevel]
        for nodeId in self.nodeDic:
            if self.getNode(nodeId).nodeClass not in ["module","text"]:
                nodeInputs = []
                nodeOutputs = []
                fullNode = []
                fullInputs = []
                fullOutputs = []
                # To clear non existent aliases
                if self.getNode(nodeId).originalId:
                    if self.getNode(nodeId).originalId not in self.nodeDic:
                        continue
                nodeInputs = self.getNode(nodeId).inputs
                nodeOutputs = self.getNode(nodeId).outputs
                fullNode = self.getNodeParentsAndAliases(nodeId)
                if fullNode:
                    for n in fullNode:
                        if n not in thisIds:
                            fullNode.remove(n)
                if fullNode:
                    if nodeInputs:
                        for i in nodeInputs:
                            fullInputs = self.getNodeParentsAndAliases(i)
                            if fullInputs:
                                for n in fullInputs:
                                    if n not in thisIds:
                                        fullInputs.remove(n)
                            if fullInputs:
                                for n in fullNode:
                                    for inp in fullInputs:
                                        if self.getNode(n).nodeInfo.showInputs and self.getNode(inp).nodeInfo.showOutputs:
                                            element = {"from":inp,"to":n}
                                            if element not in res:
                                                res.append(element)

                    if nodeOutputs:
                        for o in nodeOutputs:
                            fullOutputs = self.getNodeParentsAndAliases(o)
                            if fullOutputs:
                                for n in fullOutputs:
                                    if n not in fullOutputs:
                                        fullOutputs.remove(n)
                            if fullOutputs:
                                for n in fullNode:
                                    for out in fullOutputs:
                                        if self.getNode(out).nodeInfo.showInputs and self.getNode(n).nodeInfo.showOutputs:
                                            element = {"from":n,"to":out}
                                            if element not in res:
                                                res.append(element)

        return res


    def getNodeParentsAndAliases(self,nodeId):
        response = []
        aliases = []
        parentModulesAndAliases = []
        if nodeId not in response:
            response.append(nodeId)
        aliases = self.findNodes('originalId',nodeId)
        if aliases:
            for a in aliases:
                if a.identifier not in response:
                    response.append(a.identifier)
        if self.getNode(nodeId).isin in self.nodeDic:
            parentModulesAndAliases = self.getParentModulesWithAlias(
                self.getNode(nodeId).isin,[])
        if parentModulesAndAliases:
            for m in parentModulesAndAliases:
                if m not in response:
                    response.append(m)
        return response

    """

    def findNodes(self, prop, value):
        """Finds nodes by property/value"""
        res = []
        for k, v in self.nodeDic.items():
            if getattr(v, prop) == value:
                if(not v.system):
                    res.append(self.nodeDic[k])
        return res

    def searchNodes(self, filterOptions):
        """Search nodes using filter options """
        res = []
        res = Intellisense().search(self, filterOptions)
        return res

    def getInputs(self, nodeId):
        res = []
        if self.existNode(nodeId):
            for nodeInput in self.getNode(nodeId).inputs:
                if self.existNode(nodeInput):
                    inp = self.getNode(nodeInput)
                    res.append({
                        'id': nodeInput,
                        'name': inp.title if not inp.title is None else nodeInput,
                        'nodeClass': inp.nodeClass,
                        'moduleId': inp.moduleId
                    })

        return res

    def getOutputs(self, nodeId):
        """Get output list of a node"""
        res = []
        if self.existNode(nodeId):
            for nodeOutput in self.getNode(nodeId).outputs:
                if self.existNode(nodeOutput):
                    out = self.getNode(nodeOutput)
                    res.append({
                        'id': nodeOutput,
                        'name': out.title if not out.title is None else nodeOutput,
                        'nodeClass': out.nodeClass,
                        'moduleId': out.moduleId
                    })
        return res

    def moveNodes(self, nodeList, moduleId):
        """Move nodes to other moduleId"""
        res = []
        moduleId = self.clearId(moduleId)
        if self.existNode(moduleId):
            for nodeId in nodeList:
                if self.existNode(nodeId):
                    self.getNode(nodeId).moduleId = moduleId
                    res.append(nodeId)
        return res

    def copyNodes(self, nodeList, moduleId):
        """Copy nodes"""
        res = []
        if self.existNode(moduleId):
            try:
                self._isLoadingModel = True
                rx = r"('[^'\\]*(?:\\.[^'\\]*)*'|\"[^\"\\]*(?:\\.[^\"\\]*)*\")|\b{0}\b"
                newNodesDic = dict()

                def nodeCreator(_nodeList, _moduleId):

                    for nodeId in _nodeList:
                        nodeId = self.clearId(nodeId)
                        if self.existNode(nodeId):
                            obj = self.getNode(nodeId).toObj()

                            newId = self.getNextIdentifier(
                                f"{obj['identifier']}")
                            newNodesDic[obj['identifier']] = newId
                            obj['identifier'] = newId

                            if obj['moduleId'] == _moduleId:
                                obj['x'] = int(obj['x']) + 10
                                obj['y'] = int(obj['y']) + 10
                            else:
                                obj['moduleId'] = _moduleId
                            nodeObj = self.createNode(
                                obj['identifier'], moduleId=_moduleId)
                            nodeObj.fromObj(obj)
                            res.append(nodeObj.identifier)

                            if nodeObj.nodeClass == 'module':
                                _childrens = [
                                    node.identifier for node in self.findNodes('moduleId', nodeId)]
                                nodeCreator(_childrens, newId)

                nodeCreator(nodeList, moduleId)

                # update definitions and alias original
                for sourceNode, targetNode in newNodesDic.items():
                    newNode = self.getNode(targetNode)
                    currentDef = newNode.definition
                    if not currentDef is None and currentDef != '':
                        tmpCode = newNode.compileDef(currentDef)
                        if not tmpCode is None:
                            names = newNode.parseNames(tmpCode)
                            for node in names:
                                if node in newNodesDic:
                                    newRelatedId = newNodesDic[node]

                                    currentDef = re.sub(rx.format(node), lambda m:
                                                        (
                                        m.group(1)
                                        if m.group(1)
                                        else
                                        newRelatedId
                                    ), currentDef, 0, re.IGNORECASE)
                                    # end parse definition
                            newNode.definition = currentDef

                    # update allias
                    if not newNode.originalId is None and newNode.originalId in newNodesDic:
                        newNode.originalId = newNodesDic[newNode.originalId]

                    # regenerate
                    newNode.generateIO()

            finally:
                self._isLoadingModel = False

        return res

    def copyAsValues(self, nodeId, asNewNode=False):
        """ Copy node as values """
        if self.existNode(nodeId):
            node = self.nodeDic[nodeId]
            if node.originalId:
                return self.copyAsValues(node.originalId, asNewNode)

            result = node.result
            if asNewNode:
                newNode = self.createNode(moduleId=node.moduleId, nodeClass=node.nodeClass, x=int(
                    node.x)+40, y=int(node.y)+60)
                newNode.w = node.w
                newNode.h = node.h
                newNode.definition = node.definition
                node = newNode
            evaluator = Evaluator.createInstance(result)
            return evaluator.copyAsValues(result, self.nodeDic, node.identifier)

        return False

    def createInputNodes(self, nodeList):
        """Create input nodes"""
        res = []
        if not nodeList is None:
            for nodeId in nodeList:
                if self.existNode(nodeId):
                    firstNode = self.getNode(nodeId)
                    nodeOrig = self.getOriginalNode(nodeId)
                    inputNode = self.createNode(moduleId=firstNode.moduleId, nodeClass='formnode', x=int(
                        firstNode.x)-70, y=int(firstNode.y)+70, originalId=nodeOrig.identifier)
                    inputNode.w = 240
                    inputNode.h = 36
                    inputNode.color = nodeOrig.color
                    inputNode.definition = 'result = ' + \
                        str(nodeOrig.identifier)

                    res.append(inputNode.identifier)
        return res

    def createAlias(self, nodeList):
        """Create node alias"""
        res = []
        if not nodeList is None:
            for nodeId in nodeList:
                if self.existNode(nodeId):
                    firstNode = self.getNode(nodeId)
                    nodeOrig = self.getOriginalNode(nodeId)
                    aliasNode = self.createNode(moduleId=firstNode.moduleId, nodeClass='alias', x=int(
                        firstNode.x)+30, y=int(firstNode.y)+30, originalId=nodeOrig.identifier)
                    aliasNode.w = int(nodeOrig.w)
                    aliasNode.h = int(nodeOrig.h)
                    aliasNode.definition = 'result = ' + \
                        str(nodeOrig.identifier)
                    aliasNode.color = BaseNode.getDefaultColor(
                        nodeOrig.nodeClass) if nodeOrig.color is None else nodeOrig.color

                    res.append(aliasNode.identifier)
        return res

    def getOriginalNode(self, nodeId):
        """Get original node from an alias or input node"""
        if self.existNode(nodeId):
            nodeOrig = self.getNode(nodeId)
            if nodeOrig.originalId is None:
                return nodeOrig
            else:
                return self.getOriginalNode(nodeOrig.originalId)

    def isCalcNodes(self, nodeList):
        """Return list of Booleans. True for node is calculated otherwise False"""
        res = []
        if not nodeList is None:
            for nodeId in nodeList:
                isCalc = False
                if self.existNode(nodeId):
                    isCalc = self.getNode(nodeId).isCalc
                res.append(isCalc)
        return res

    def exportModule(self, moduleId, filename):
        """Export module to file"""
        _moduleIOEngine = IOModule(self)
        return _moduleIOEngine.exportModule(moduleId, filename)

    def importModule(self, moduleId, filename, importType):
        """Import module from file"""
        _moduleIOEngine = IOModule(self)
        return _moduleIOEngine.importModule(moduleId, filename, importType)

    def exportFlatNode(self, nodeId, numberFormat, columnFormat, fileName):
        """Export node values in flat format"""
        if self.existNode(nodeId):
            evaluator = Evaluator.createInstance(self.nodeDic[nodeId].result)
            return evaluator.exportFlatNode(self.nodeDic, nodeId, numberFormat, columnFormat, fileName)
        return False

    def dumpNodeToFile(self, nodeId, fileName):
        """Dump current node value to file"""
        if self.existNode(nodeId):
            evaluator = Evaluator.createInstance(self.nodeDic[nodeId].result)
            return evaluator.dumpNodeToFile(self.nodeDic, nodeId, fileName)
        return False

    def saveModel(self, fileName=None):
        """Save model. If fileName is specified, then save to fileName, else return string of ppl """

        # update used libraries
        self.modelProp['libs'] = self._get_used_libraries()

        toSave = {
            'modelProp': self.modelProp,
            'nodeList': []
        }

        for k, v in self.nodeDic.items():
            if(not v.system):
                toSave['nodeList'].append(v.toObj())

        if fileName:
            # Save existing model
            if os.path.isfile(fileName):
                filename_to_save = f'{fileName}.tmp#'
                filename_old = f'{fileName}.old#'

                # remove old .tmp file
                if os.path.isfile(filename_to_save):
                    os.remove(filename_to_save)
                # remove old .old file
                if os.path.isfile(filename_old):
                    os.remove(filename_old)

                with open(filename_to_save, 'w') as f:
                    f.write(jsonpickle.encode(toSave))
                    f.close()

                # move old ppl to filename_old
                os.rename(fileName, filename_old)
                # move filename_to_save to ppl
                os.rename(filename_to_save, fileName)

                if os.path.isfile(filename_old):
                    new_size = os.path.getsize(fileName)
                    old_size = os.path.getsize(filename_old)
                    if new_size/old_size > 0.8:
                        # remove old if new file is greater
                        os.remove(filename_old)
                    else:
                        os.rename(
                            filename_old, f"{fileName}-{datetime.datetime.today().strftime('%Y%m%d-%H%M%S')}.old")

            # Save new model
            else:
                with open(fileName, 'w') as f:
                    f.write(jsonpickle.encode(toSave))
                    f.close()
        return jsonpickle.encode(toSave)

    def openModel(self, fileName=None, textModel=None):
        """Open model.
        If fileName is especified then open from fileName, else open from textModel text """

        self.release()

        opened = {}
        if textModel:
            opened = jsonpickle.decode(textModel)
        else:
            with open(fileName, 'r') as f:
                opened = jsonpickle.decode(f.read())
                f.close()

        self._modelProp = opened['modelProp']

        self._isLoadingModel = True

        # create base model node
        multiplier = 1
        hasBaseNode = False
        for obj in opened['nodeList']:
            if obj['moduleId'] == '_model_':
                node = self.createNode(obj['identifier'], moduleId='_model_')
                # multiplier for old models
                if obj['w'] == 50 and obj['h'] == 25:
                    # the model is an old model
                    multiplier = 2
                obj['w'] = str(int(obj['w']) * multiplier)
                obj['h'] = str(int(obj['h']) * multiplier)
                node.fromObj(obj)
                self._modelNode = node
                hasBaseNode = True
                break

        if not hasBaseNode:
            self.initialize()

        # creating system nodes
        self.createSystemNodes(fileName)

        rootModelId = self.modelNode.identifier

        # create nodes
        for obj in opened['nodeList']:
            if obj['moduleId'] != '_model_' and obj['identifier']:
                if obj['nodeClass'] in ['alias', 'formnode']:
                    if hasattr(obj, 'originalId'):
                        node = self.createNode(
                            obj['identifier'], moduleId=rootModelId, originalId=obj['originalId'])
                    else:
                        index = obj['definition'].find('=')
                        originalId = obj['definition'][index+1:].strip()
                        node = self.createNode(
                            obj['identifier'], moduleId=rootModelId, originalId=originalId)
                else:
                    node = self.createNode(
                        obj['identifier'], moduleId=rootModelId)

                obj['w'] = str(int(obj['w']) * multiplier)
                obj['h'] = str(int(obj['h']) * multiplier)
                node.fromObj(obj)
                node = None

        [self.nodeDic[nod].generateIO() for nod in self.nodeDic]

        opened = None
        self._isLoadingModel = False

        # check models library
        self.ensureModelLibraries()

        # apply backward compatibility
        self.applyBackwardCompatibility()

        # evaluate nodes on start
        try:
            for key in self.nodeDic:
                if self.nodeDic[key] and self.nodeDic[key].evaluateOnStart:
                    _dummy = self.nodeDic[key].result
        except Exception as ex:
            print(str(ex))
            # TODO: send via channels msg to client
            pass

        return True

    def closeModel(self):
        """Close model"""
        self.release()

    def getCustomImports(self):
        """Return object with custom imported python modules."""
        if self._customImports is None:
            self._buildCustomImports()
        return self._customImports

    def _buildCustomImports(self):
        """Build custom imports"""

        self._customImports = Model.DEFAULT_IMPORTS.copy()

        # add pyplan funcions
        self._customImports["pp"] = PyplanFunctions(self)
        # for backward compatibility
        self._customImports["selector"] = self._customImports["pp"].selector

        # support old 'default imports' node
        if self.existNode('imports'):
            import_node = self.getNode('imports')
            import_dic = import_node.result
            for key in import_dic:
                if not key in self._customImports:
                    self._customImports[key] = import_dic[key]

    def createSystemNodes(self, fileName):
        """Create system nodes"""
        # current path
        systemPathNode = self.createNode(
            identifier='current_path', moduleId=self.modelNode.identifier)

        path = os.path.abspath(
            fileName[0:fileName.rfind(os.path.sep)]) + os.path.sep
        if self.isLinux():
            path = fileName[:fileName.rfind('/')] + '/'
            self.createSymlinks(path)
        else:
            path = path.replace('\\', '\\\\')

        systemPathNode.system = True
        systemPathNode.definition = 'result="""' + str(path) + '"""'
        os.chdir(str(path))

        node = self.createNode(identifier='pyplan_user',
                               moduleId=self.modelNode.identifier)
        node.system = True

        node = self.createNode(identifier='cub_refresh',
                               moduleId=self.modelNode.identifier)
        node.system = True
        node = self.createNode(identifier='pyplan_refresh',
                               moduleId=self.modelNode.identifier)
        node.system = True

        node = self.createNode(
            identifier='_scenario_', moduleId=self.modelNode.identifier, nodeClass='index')
        node.system = True
        node.title = 'Scenario'
        node.definition = "result = pp.index(['Current'])"

        node = self.createNode(identifier='task_log_endpoint',
                               moduleId=self.modelNode.identifier, nodeClass='variable')
        node.system = True
        node.title = 'TaskLog endpoint'
        node.definition = "result = ''"

    def createSymlinks(self, fileName):
        if os.getenv('PYPLAN_IDE', '0') == '1' or os.getenv('ENGINE_MODE', '') in ['fixed', 'local']:
            return

        # Add user or public path to system paths
        path_sep = os.path.sep
        path_arr = fileName[:fileName.rfind(path_sep)].split(path_sep)

        dest_path = path_sep.join(
            path_arr[:6 if path_arr[4] == 'Workgroups' else 5])

        # Get python folder path
        python_folder = f'python{sys.version[:3]}'
        try:
            folder_list = os.listdir(os.path.join(dest_path, '.venv', 'lib'))
            python_folder = folder_list[len(folder_list)-1]
        except Exception as ex:
            pass

        # Add user/public library to system paths
        user_lib_path = os.path.join(
            dest_path, '.venv', 'lib', python_folder, 'site-packages')
        venv_path = os.path.join('/venv', 'lib', 'python3.7', 'site-packages')

        # Check Write Access to user_lib_path
        if os.access(user_lib_path, os.W_OK):
            if not os.path.isdir(user_lib_path):
                os.makedirs(user_lib_path, exist_ok=True)

            # copy base venv folders
            os.system(f'cp -r -u {venv_path}-bkp/* {user_lib_path}')

            # create symlink from user /public site-package
            os.system(f'rm -rf {venv_path}')
            os.system(f'ln -s -f {user_lib_path} {venv_path}')

    def applyBackwardCompatibility(self):
        # update old selector definition
        if self.existNode("selector"):
            node = self.getNode("selector")
            if node.nodeClass == "function" and "from pyplan_engine.classes.PyplanFunctions import Selector" in node.definition:
                node.definition = "result = pp.selector"

        # check for cubepy in imports node
        if self.existNode("imports"):
            node = self.getNode("imports")
            if "cubepy" in node.definition and not "cubepy" in sys.modules:
                sys.modules["cubepy"] = cubepy


    def isLinux(self):
        if platform == 'linux' or platform == 'linux2' or platform == 'darwin':
            return True
        else:
            return False

    # profileNode
    def profileNode(self, nodeId):
        """Perform profile of an node"""
        if self.getNode(nodeId).originalId is not None:
            nodeId = self.getNode(nodeId).originalId
        profile = self.getNode(nodeId).profileNode(
            [], [], self.getNode(nodeId).evaluationVersion, nodeId)

        for node in profile:
            node['calcTime'] = node['evaluationTime'] if node['evaluationTime'] > 0 else 0

        # Fix acumulated time
        total_time = 0
        for nn in reversed(range(len(profile))):
            node = profile[nn]
            total_time = total_time + node['calcTime']
            node['evaluationTime'] = total_time

        return jsonpickle.encode(profile)

    def evaluate(self, definition, params=None, returnEvaluateTime=False):
        """Evaluate expression"""
        res = None
        evalNode = BaseNode(
            model=self, identifier='__evalnode__', nodeClass='variable')
        evalNode._definition = definition
        evalNode.calculate(params)
        res = evalNode.result
        evaluateTime = evalNode.lastEvaluationTime
        evalNode.release()
        evalNode = None
        if returnEvaluateTime:
            return res, evaluateTime
        return res

    def callFunction(self, nodeId, params=None):
        """Call node function with params"""
        res = None
        if self.existNode(nodeId):
            nodeFn = self.getNode(nodeId).result
            if params is None:
                res = nodeFn()
            else:
                res = nodeFn(**params)
        return res

    def getIdentifierByNode(self, result):
        """Return Identifier of node searching by your result"""
        res = ''
        for nodeId in self.nodeDic:
            if self.nodeDic[nodeId].isCalc:
                if self.nodeDic[nodeId].result is result:
                    res = nodeId
                    break
        return res

    def loadScenario(self, nodeId, scenarioData):
        """ Load and fill scenarioDic """
        res = False
        if not scenarioData is None:
            if self.existNode(nodeId):

                scenarioResult = []
                scenarioNames = []

                scenList = str(scenarioData).split('##')
                for scenItem in scenList:
                    arr = str(scenItem).split('||')
                    if len(arr) == 3:
                        scenarioName = arr[1]
                        fileName = arr[2]
                        _result = None
                        if arr[0] == '-1':  # current
                            _result = self.getNode(nodeId).result
                        else:
                            nodeDef = ''
                            with open(fileName, 'r') as f:
                                nodeDef = f.read()
                                f.close()
                            _result = self.evaluate(nodeDef)
                        scenarioResult.append(_result)
                        scenarioNames.append(scenarioName)

                # fill scenario node
                scenarioIndex = self.getNode('_scenario_')
                scenarioIndex.definition = "result = pp.index(['" + "','".join(
                    scenarioNames) + "'])"

                if len(scenarioResult) > 0:
                    finalResult = None

                    if isinstance(scenarioResult[0], xr.DataArray):
                        finalResult = xr.concat(
                            scenarioResult, scenarioIndex.result.dataArray)
                    else:
                        finalResult = xr.DataArray(
                            scenarioResult, scenarioIndex.result.coord)

                    self._scenarioDic[nodeId] = finalResult
                    res = True

        return res

    def unloadScenario(self):
        """Unload all scenarios and clean calculated nodes in the scenario """
        for key in self._scenarioDic:
            self._scenarioDic[key] = None

        self._scenarioDic = dict()
        scenarioIndex = self.getNode('_scenario_')
        scenarioIndex.definition = "result = pp.index(['Current'])"

    def geoUnclusterData(self, nodeId, rowIndex, attIndex, latField='latitude', lngField='longitude', geoField='geoField', labelField='labelField', sizeField='sizeField', colorField='colorField', iconField='iconField'):
        """get unclusted data for geo representation."""
        if self.existNode(nodeId):
            result = self.nodeDic[nodeId].result
            if not result is None:
                self.evaluationVersion += 1
                evaluator = Evaluator.createInstance(result)
                return evaluator.geoUnclusterData(result, self.nodeDic, nodeId, rowIndex, attIndex, latField, lngField, geoField, labelField, sizeField, colorField, iconField)
        else:
            return ''

    def callWizard(self, param):
        """ Start and call wizard toolbar """
        obj = jsonpickle.decode(param)
        key = obj['wizard']
        action = obj['action']
        params = obj['params']

        if self._wizard is None or self._wizard.code != key:
            self._wizard = self._getWizzard(key)

        toCall = getattr(self._wizard, action)

        return toCall(self, params)

    def createNewModel(self, modelFile=None, modelName=None):
        self.release()
        self._modelProp = {}
        self._isLoadingModel = True
        self.initialize(modelName)
        self.createSystemNodes(modelFile)
        self._isLoadingModel = False
        if modelFile:
            self.saveModel(modelFile)
        return True

    def _getWizzard(self, key):
        key = key.lower()
        if key == 'calculatedfield':
            return CalculatedField.Wizard()
        elif key == 'selectcolumns':
            return SelectColumns.Wizard()
        elif key == 'selectrows':
            return SelectRows.Wizard()
        elif key == 'sourcecsv':
            return sourcecsv.Wizard()
        elif key == 'dataframeindex':
            return DataframeIndex.Wizard()
        elif key == 'dataframegroupby':
            return DataframeGroupby.Wizard()
        elif key == 'dataarrayfrompandas':
            return DataarrayFromPandas.Wizard()
        elif key == 'inputtable':
            return InputTable.Wizard()
        elif key == 'createindex':
            return CreateIndex.Wizard()
        elif key == 'indexfrompandas':
            return IndexFromPandas.Wizard()
        elif key == 'dataarrayfilter':
            return DataArrayFilter.Wizard()
        

    def getSystemResources(self, onlyMemory=False):
        """Return current system resources"""

        def _read_int(file):
            data = 0
            with open(file, 'r') as f:
                data = int(f.read())
                f.close()
            return data

        def _read_cache():
            data = 0
            with open('/sys/fs/cgroup/memory/memory.stat', 'r') as f:
                line = f.readline()
                data = int(str(line).split(' ')[1])
                f.close()
            return data

        mem_limit = _read_int(
            '/sys/fs/cgroup/memory/memory.limit_in_bytes') / 1024/1024/1024
        if mem_limit > 200:  # bug for container whitout limit
            total_host = ''
            with open('/proc/meminfo', 'r') as f:
                line1 = f.readline()
                total_host = str(line1).split(' ')[-2:-1][0]
                mem_limit = int(total_host) / 1024/1024

        mem_used = (_read_int(
            '/sys/fs/cgroup/memory/memory.usage_in_bytes') - _read_cache()) / 1024/1024/1024

        # get cpu usage

        if onlyMemory:
            return {
                'totalMemory': mem_limit,
                'usedMemory': mem_used
            }
        else:

            time_1 = datetime.datetime.now()
            cpu_time_1 = _read_int('/sys/fs/cgroup/cpu/cpuacct.usage')
            sleep(0.3)
            time_2 = datetime.datetime.now()
            cpu_time_2 = _read_int('/sys/fs/cgroup/cpu/cpuacct.usage')
            delta_time = (time_2 - time_1).microseconds * 10
            used_cpu = (cpu_time_2 - cpu_time_1) / delta_time
            used_cpu = used_cpu if used_cpu < 100 else 100

            current_node = self._currentProcessingNode
            if self.existNode(current_node):
                node = self.getNode(current_node)
                if node.title:
                    current_node = f'{node.title} ({current_node})'

            return {
                'totalMemory': mem_limit,
                'usedMemory': mem_used,
                'usedCPU': used_cpu,
                'pid': self.getPID(),
                'currentNode': current_node
            }

    def ensureModelLibraries(self):
        """Ensure that all model libs are installed"""
        try:
            if not 'libs' in self.modelProp:
                self.modelProp['libs'] = []
            modelLibs = self.modelProp['libs']
            # get current installed libs
            installed_libs_str = self.listInstalledLibraries()
            installed_libs = jsonpickle.decode(installed_libs_str)
            installed_libs_dic = dict()
            for installed_lib in installed_libs:
                installed_libs_dic[installed_lib['name']
                                   ] = installed_lib['version']

            to_install = ''
            for lib in modelLibs:
                if not lib['name'] in installed_libs_dic:
                    to_install += f" {lib['name']}=={lib['version']}"
                # TODO: if exists, check version and send message via channels

            if to_install != '':
                self._installLibrary(to_install)

        except Exception as ex:
            print(f'Error checking libraries. {ex}')
            # TODO: send to client via channels

    def _installLibrary(self, lib):

        cmd = f'pip install {lib}'

        # If there are proxy configurations, use them to install from pip
        http_proxy = os.getenv('PYPLAN_HTTP_PROXY')
        https_proxy = os.getenv('PYPLAN_HTTPS_PROXY')
        if http_proxy:
            cmd += f' --proxy {http_proxy}'
        elif https_proxy:
            cmd += f' --proxy {https_proxy}'

        p = subprocess.Popen(split(cmd), stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, universal_newlines=True)
        nn = 0
        while p.stdout is not None and nn < 240:
            # TODO: show feedback to ide using channels
            line = p.stdout.readline()
            if not line:
                p.stdout.flush()
                aa, err = p.communicate()
                if err:
                    self._currentInstallProgress.append(err.rstrip('\n'))
                break
            sleep(1)
            nn += 1
            self._currentInstallProgress.append(line.rstrip('\n'))

        importlib.invalidate_caches()

    def installLibrary(self, lib, target):
        """install python library"""
        self._currentInstallProgress = []
        thread = threading.Thread(target=self._installLibrary, args=(lib,))
        thread.start()
        return 'ok'

    def _get_used_libraries(self):
        try:
            _regex = re.compile(r'(^import\s.*)|(^from\s.*)', re.M)
            _temp_imports = []
            _imports = dict()
            _used_libraries = []

            # retrieve all import statements
            for node_id in self.nodeDic:
                _def = self.nodeDic[node_id].definition
                if _def:
                    _finditer = re.finditer(_regex, _def)
                    for match in _finditer:
                        if match.group() not in _temp_imports:
                            _temp_imports.append(match.group())

            # filter imports (make distinct, check type)
            # name == pypi name
            for _element in _temp_imports:
                if _element[:6] == 'import':
                    # import
                    if _element[7:].find(',') != -1:
                        # multiple imports in one line
                        for _el in _element[7:].split(','):
                            _imports[self._check_import_function(_el.strip())] = {'import_name': self._check_import_function(
                                _el.strip()), 'import_type': 'import', 'name': None, 'version': None}
                    elif _element[7:].find(' as ') != -1:
                        # import has an alias
                        _el = _element[7:_element.find(' as ')].strip()
                        _imports[self._check_import_function(_el)] = {'import_name': self._check_import_function(
                            _el), 'import_type': 'import', 'name': None, 'version': None}
                    else:
                        # full import
                        _el = _element[7:].strip()
                        _imports[self._check_import_function(_el)] = {'import_name': self._check_import_function(
                            _el), 'import_type': 'import', 'name': None, 'version': None}
                elif _element[:4] == 'from':
                    # from
                    if _element[5:].find(' import ') != -1:
                        _el = _element[5:_element.find(' import ')].strip()
                        _imports[self._check_import_function(_el)] = {'import_name': self._check_import_function(
                            _el), 'import_type': 'from', 'name': None, 'version': None}
                    elif self.ws:
                        self.ws.sendMsg(_element, 'Could not find import from')
                elif self.ws:
                    self.ws.sendMsg(_element, 'Element not recognized')

            _installed_libs = self._check_installed_libraries()
            for key in _imports.keys():
                for lib in _installed_libs:
                    if lib == _imports[key]['import_name']:
                        _imports[key].update(_installed_libs[lib])
                if _imports[key]['name'] != None:
                    _used_libraries.append(_imports[key])

            return _used_libraries
        except Exception as ex:
            if self.ws:
                self.ws.sendMsg(str(ex), 'Error getting used libraries')
            return []

    def _check_import_function(self, _element):
        if _element.find('.') != -1:
            # the element has a function been imported
            return _element[:_element.find('.')].strip()
        else:
            return _element

    def _check_installed_libraries(self):
        try:
            installed_libraries = dict()
            dirs = sys.path if os.getenv('PYPLAN_IDE') else [
                '/venv/lib64/python3.7/site-packages/']
            for _dir in dirs:
                installed_modules = dict()
                if '.zip' not in _dir:
                    for folder in os.listdir(_dir):
                        # get dist.info folders
                        if '.dist-info' in folder or '.egg-info' in folder:
                            top_level_file = os.path.join(
                                _dir, folder, 'top_level.txt')
                            if os.path.isfile(top_level_file):
                                top_level = str(
                                    open(top_level_file).read()).replace('\n', '')
                                metadata_file = os.path.join(
                                    _dir, folder, 'METADATA' if '.dist-info' in folder else 'PKG-INFO')
                                if os.path.isfile(metadata_file):
                                    metadata = str(open(metadata_file).read())
                                    metadata_arr = metadata.split('\n')
                                    for metadata_item in metadata_arr:
                                        if str(metadata_item).startswith('Name: '):
                                            pypi_name = str(
                                                metadata_item).split(' ')[1]
                                        elif str(metadata_item).startswith('Version: '):
                                            pypi_version = str(
                                                metadata_item).split(' ')[1]
                                installed_modules[top_level] = {
                                    'name': pypi_name, 'version': pypi_version}
                installed_libraries.update(installed_modules)
            return installed_libraries
        except Exception as ex:
            if self.ws:
                self.ws.sendMsg(str(ex), 'Error checking installed libraries')
            return {}

    def listInstalledLibraries(self):
        cmd = 'pip list -v --disable-pip-version-check --format=json'
        popen = subprocess.Popen(
            split(cmd), stdout=subprocess.PIPE, universal_newlines=True)

        stdout, stderr = popen.communicate()
        if stderr:
            raise ValueError(
                f'Error listing installed libraries: {str(stderr)}')

        return stdout

    def uninstallLibrary(self, lib, target):
        """Uninstall python library"""
        # cmd = f'pip uninstall -y {lib}'
        # We cant use pip uninstall so we do a dirty workaround
        cmd = f"find {target} -name '*{lib.replace('-','_')}*' -exec rm -rf {{}} \;"
        popen = subprocess.Popen(
            split(cmd), stdout=subprocess.PIPE, universal_newlines=True)
        stdout, stderr = popen.communicate()
        importlib.invalidate_caches()
        if stderr:
            raise ValueError(f'Error uninstalling library: {str(stderr)}')

        return stdout

    def getInstallProgress(self, from_line):
        """Return install library progress"""
        from_line = int(from_line)
        if len(self._currentInstallProgress) == 0 or from_line > len(self._currentInstallProgress):
            return []
        else:
            return self._currentInstallProgress[from_line:]

    def setNodeIdFromTitle(self, node_id):
        """Generate node id from node identifier """
        res = {'node_id': node_id}
        model_props = self.getModelProperties()
        if (not 'changeIdentifier' in model_props or model_props['changeIdentifier'] == '1') and self.existNode(node_id):
            node = self.nodeDic[node_id]
            new_id = ''
            try:
                if node.title:
                    new_id = self._removeDiacritics(node.title)
            except Exception as ex:
                raise ValueError(f'Error finding node title: {str(ex)}')

            if new_id:
                if self.existNode(new_id):
                    new_id = self.getNextIdentifier(new_id)
                node.identifier = new_id
                res['node_id'] = new_id
        return res

    def _removeDiacritics(self, text):
        """Removes all diacritic marks from the given string"""
        norm_txt = unicodedata.normalize('NFD', text)
        shaved = ''.join(c for c in norm_txt if not unicodedata.combining(c))
        # remove accents and other diacritics, replace spaces with '_' because identifiers can't have spaces
        no_spaces = unicodedata.normalize(
            'NFC', shaved).lower().replace(' ', '_')
        final_text = no_spaces
        # only allow [a-z], [0-9] and _
        p = re.compile('[a-z0-9_]+')
        for i in range(0, len(no_spaces)):
            if not (p.match(no_spaces[i])):
                final_text = final_text[:i] + '_' + final_text[i+1:]
        # i the first char is not a-z then replaceit (all identifiers must start with a letter)
        p2 = re.compile('[a-z]+')
        if not p2.match(final_text[0]):
            final_text = 'a' + final_text[1:]
        return final_text



from pyplan_core.classes.dynamics.BaseDynamic import BaseDynamic
from pyplan_core.cubepy.Helpers import Helpers
import numpy as np
import re
import xarray as xr
import datetime as dt


class PureXArrayDynamic(BaseDynamic):

    def circularEval(self, node, params):
        """
        Used for execute nodes with circular reference (pp.dynamic)
        """

        dynamicVars = params["dynamicVars"]
        dynamicIndex = params["dynamicIndex"]
        nodesInCyclic = params["nodesInCyclic"]
        initialValues = params["initialValues"]
        shift = params["shift"]

        evaluate = node.model.evaluate

        if node.model.debugMode:
            for circular_node_id in nodesInCyclic:
                circular_node = node.model.getNode(circular_node_id)
                if not circular_node is None:
                    circular_node.sendStartCalcNode(fromDynamic=True)

        # create nodes array
        cyclicNodes = []

        try:
            node.model.inCyclicEvaluate = True
            for nodeId in nodesInCyclic:
                _nodeObj = node.model.getNode(nodeId)
                cyclic_item = {
                    "node": _nodeObj,
                    "initialize": self.generateInitDef(node, _nodeObj.bypassCircularEvaluator().result, dynamicIndex),
                    "calcTime": _nodeObj.lastEvaluationTime
                }
                startTime = dt.datetime.now()
                cyclic_item["loopDefinition"] = self.generateLoopDef(
                    node, _nodeObj.definition, nodesInCyclic)
                endTime = dt.datetime.now()
                cyclic_item["calcTime"] += (endTime -
                                            startTime).total_seconds()

                cyclicNodes.append(cyclic_item)
        except Exception as e:
            raise e
        finally:
            node.model.inCyclicEvaluate = False

        cyclicDic = {}

        # initialice var dictionary
        for _node in cyclicNodes:
            _id = _node["node"].identifier

            evaluate_node_time = 0
            evaluate_initial_params_time = 0

            cyclicDic[_id], evaluate_node_time = evaluate(
                _node["initialize"], returnEvaluateTime=True)

            if not initialValues is None and _id in initialValues:
                initial_result, evaluate_initial_params_time = evaluate(
                    initialValues[_id], returnEvaluateTime=True)
                cyclicDic[_id] = cyclicDic[_id] + initial_result
            # check align
            if cyclicDic[_id].dims[0] != dynamicIndex.name:
                # move dynamic index to top
                list_dims = list(cyclicDic[_id].dims)
                list_dims.remove(dynamicIndex.name)
                new_tuple = (dynamicIndex.name,) + tuple(list_dims)
                cyclicDic[_id] = cyclicDic[_id].transpose(
                    *new_tuple, transpose_coords=True)

            _node["calcTime"] = _node["calcTime"] + \
                evaluate_node_time + evaluate_initial_params_time

        # initialice vars in t-1
        for _var in dynamicVars:
            _key = "__" + _var + "_t"
            cyclicDic[_key] = cyclicDic[_var].sum(dynamicIndex.name) * 0

        # loop over index
        cyclicParams = None
        # for item in dynamicIndex:

        theRange = range(0, len(dynamicIndex.values))
        initialCount = shift*-1
        reverseMode = False
        if shift > 0:
            theRange = range(len(dynamicIndex.values)-1, -1, -1)
            initialCount = len(dynamicIndex.values)-1-shift
            reverseMode = True

        for nn in theRange:
            item = dynamicIndex.values[nn]

            # load params
            cyclicParams = {
                "item": item,
                "cyclicDic": cyclicDic,
                "dynamicIndex": dynamicIndex,
                "self": self
            }

            __initialValues = None
            # loop over variables
            for _node in cyclicNodes:

                _id = _node["node"].identifier
                node.model.currentProcessingNode(_id)
                cyclicParams["cyclicDic"] = cyclicDic

                evaluate_node_time = 0
                evaluate_initial_params_time = 0

                start_extra_process_time = None

                # execute vars
                if (_id in initialValues) and ((nn < initialCount and (not reverseMode)) or (nn > initialCount and reverseMode)):
                    # use initial values
                    _resultNode, evaluate_node_time = evaluate(
                        _node["loopDefinition"], cyclicParams, True)
                    _initialValues, evaluate_initial_params_time = evaluate(
                        initialValues[_id], returnEvaluateTime=True)
                    _finalNode = None
                    start_extra_process_time = dt.datetime.now()
                    if isinstance(_initialValues, xr.DataArray):
                        _finalNode = self._tryFilter(
                            _resultNode, dynamicIndex, item) + self._tryFilter(_initialValues, dynamicIndex, item)
                    else:
                        _finalNode = self._tryFilter(
                            _resultNode, dynamicIndex, item) + _initialValues

                    try:
                        cyclicDic[_id].loc[{dynamicIndex.name: slice(
                            item, item)}] = _finalNode.values
                    except Exception as ex:
                        list_dims = list(cyclicDic[_id].dims)
                        list_dims.remove(dynamicIndex.name)
                        cyclicDic[_id].loc[{dynamicIndex.name: slice(
                            item, item)}] = _finalNode.transpose(*list_dims, transpose_coords=True).values

                else:
                    # dont use use initial values
                    _resultNode, evaluate_node_time = evaluate(
                        _node["loopDefinition"], cyclicParams, True)
                    _finalNode = self._tryFilter(
                        _resultNode, dynamicIndex, item)

                    start_extra_process_time = dt.datetime.now()
                    try:
                        cyclicDic[_id].loc[{
                            dynamicIndex.name: slice(item, item)}] = _finalNode.values
                    except Exception as ex:
                        list_dims = list(cyclicDic[_id].dims)
                        list_dims.remove(dynamicIndex.name)
                        cyclicDic[_id].loc[{dynamicIndex.name: slice(
                            item, item)}] = _finalNode.transpose(*list_dims, transpose_coords=True).values

                _node["calcTime"] += evaluate_node_time + \
                    evaluate_initial_params_time
                if not start_extra_process_time is None:
                    end_extra_process_time = dt.datetime.now()
                    _node["calcTime"] += (end_extra_process_time -
                                          start_extra_process_time).total_seconds()

            # set dynamicVar
            if (not reverseMode and (nn+1) < initialCount) or (reverseMode and (nn-1) > initialCount):
                # do not update t- vars
                pass
            else:
                for _var in dynamicVars:
                    _key = "__" + _var + "_t"
                    if reverseMode:
                        cyclicDic[_key] = self._tryFilter(
                            cyclicDic[_var], dynamicIndex, dynamicIndex.values[nn+shift-1])
                    else:
                        cyclicDic[_key] = self._tryFilter(
                            cyclicDic[_var], dynamicIndex, dynamicIndex.values[nn-initialCount+1])

        # set result
        for _node in cyclicNodes:
            _id = _node["node"].identifier
            _node["node"]._result = cyclicDic[_id]
            _node["node"].__resultMemory = Helpers.getResultSize(
                _node["node"]._result)
            _node["node"]._isCalc = True
            _node["node"].lastEvaluationTime = _node["calcTime"]
            _node["node"].evaluationVersion = node.model.evaluationVersion

        if node.model.debugMode:
            for circular_node_id in nodesInCyclic:
                circular_node = node.model.getNode(circular_node_id)
                if not circular_node is None:
                    circular_node.sendEndCalcNode(fromDynamic=True)

        evaluate = None
        model = None
        cyclicDic = None
        cyclicParams = None

        return "ok"

    def generateLoopDef(self, node, nodeDefinition, cyclicVariables):
        """
        Return definition used for circular evaluator
        """
        _def = self.clearCircularDependency(
            nodeDefinition, "cyclicDic['__##node##_t']")
        finalDef = _def
        tmpCode = compile(_def, '<string>', 'exec')
        names = node.parseNames(tmpCode)
        rx = r"('[^'\\]*(?:\\.[^'\\]*)*'|\"[^\"\\]*(?:\\.[^\"\\]*)*\")|\b{0}\b"
        for nodeId in names:
            if node._model.existNode(node._model.clearId(nodeId)):
                if nodeId in cyclicVariables:
                    finalDef = re.sub(rx.format(nodeId), lambda m:
                                      (
                        m.group(1)
                        if m.group(1)
                        else
                        (
                            "cyclicDic['"+nodeId+"']"
                            if (m.endpos > m.regs[0][1]+5) and ((m.string[m.regs[0][1]:m.regs[0][1]+5] == '.node') or (m.string[m.regs[0][1]:m.regs[0][1]+8] == '.timeit('))
                            else
                            (nodeId
                             if (m.string[m.regs[0][0]-1:m.regs[0][0]+len(nodeId)] == ('.'+nodeId))
                             else "self._tryFilter( cyclicDic['"+nodeId+"'],dynamicIndex,item) "
                             )
                        )
                    ), finalDef, 0, re.IGNORECASE)

        return finalDef

    def generateInitDef(self, node, nodeCube, dynamicIndex):
        """
        Return definition used for initialice vars in circular evaluator
        """
        if isinstance(nodeCube, xr.DataArray):
            _list = list(nodeCube.dims[:])
            if not dynamicIndex.name in _list:
                _list.append(dynamicIndex.name)
            _dims = ','.join(_list)
            _def = f"result = pp.create_dataarray(0.,[{_dims}])"

            """
            _data = np.full(nodeCube.shape, 0. )
            np.set_printoptions(threshold = np.prod(_data.shape))
            _data = np.array2string(_data, separator=",", precision=20 , formatter={'float_kind':lambda x: repr(x)}).replace('\n','')

            _def = f"result = xr.DataArray({_data},{_dims})"
            """
            return _def
        else:
            return f"result = pp.create_dataarray(0.,[{dynamicIndex.name}])"
            """
            _data = np.full(dynamicIndex.shape, 0. )
            np.set_printoptions(threshold = np.prod(_data.shape))
            _data = np.array2string(_data, separator=",", precision=20 , formatter={'float_kind':lambda x: repr(x)}).replace('\n','')
            return f"result = xr.DataArray({_data}, [{dynamicIndex.name}])"
            """

    def generateCircularParameters(self, node, nodeList):
        """
        Generate paremters for call to circularEval
        """
        dynamicVars = []
        dynamicIndex = None
        nodesInCyclic = []  # nodeList TODO: Determinar orden de nodos
        initialValues = {}
        shift = -1

        for _nodeId in nodeList:
            if node.model.existNode(_nodeId):
                _def = node.model.getNode(_nodeId).definition
                if "dynamic(" in _def:

                    _startPos = (_def.find("pp.dynamic(") + 11) if "pp.dynamic" in _def else (_def.find("dynamic(") + 8 )
                    _endPos = _def.find(")", _startPos)

                    # dynamicVars = _def[_startPos:_endPos] # cc,time,-1
                    _arr = str(_def[_startPos:_endPos]).split(",")
                    _vart_1 = _arr[0].strip()
                    if not _vart_1 in dynamicVars:
                        dynamicVars.append(_vart_1)
                    dynamicIndex = node.model.getNode(_arr[1].strip()).result
                    shift = int(_arr[2])
                    if len(_arr) > 3:
                        initialValues[_nodeId] = "result = " + _arr[3].strip()

        #nodesInCyclic =  sort

        _graph = {}

        for _nodeId in nodeList:
            if node.model.existNode(_nodeId):
                _graph[_nodeId] = []
                for _outputId in node.model.getNode(_nodeId).ioEngine.outputs:
                    if _outputId in nodeList:
                        if 'dynamic('+_nodeId+',' not in str(node.model.getNode(_outputId).definition).replace(" ", ""):
                            _graph[_nodeId].append(_outputId)

        nodesInCyclic = Helpers.dfs_topsort(_graph)

        return {
            "dynamicVars": dynamicVars,
            "dynamicIndex": dynamicIndex,
            "nodesInCyclic": nodesInCyclic,
            "initialValues": initialValues,
            "shift": shift
        }

    def clearCircularDependency(self, stringDef, replaceWith="0"):
        """ Replaces pp.dynamic(x,y,z) for the desired replaceWith param
        """
        response = stringDef
        initialIndex = -1
        startIndex = -1
        finalIndex = -1
        toReplace = ''
        initialIndex = stringDef.find('pp.dynamic(')
        if initialIndex != -1:
            startIndex = initialIndex
            initialIndex = initialIndex + len('pp.dynamic(')
            if len(stringDef) > initialIndex:
                finalIndex = stringDef[initialIndex+1:].find(')')
        else:
            initialIndex = stringDef.find('dynamic(')
            if initialIndex != -1:
                startIndex = initialIndex
                initialIndex = initialIndex + len('dynamic(')
                if len(stringDef) > initialIndex:
                    finalIndex = stringDef[initialIndex+1:].find(')')

        if initialIndex != -1 and finalIndex != -1:
            toReplace = stringDef[startIndex:initialIndex + finalIndex + 2]
            if "cyclicDic" in replaceWith:
                nodeInT1 = toReplace.split(",")[0]
                nodeInT1 = nodeInT1[(nodeInT1.find("(")+1):]
                replaceWith = replaceWith.replace("##node##", nodeInT1.strip())
            response = stringDef.replace(toReplace, replaceWith)

        return response

    def _tryFilter(self, array, dim, value):
        try:
            _dic = {
                dim.name: value
            }
            return array.sel(_dic, drop=True)
        except Exception as ex:
            return array



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import jsonpickle


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "SelectColumns"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            currentDef = model.getNode(nodeId).definition
            newDef = self.getLastDefinition(currentDef)
            newDef = newDef + "\n# Selected columns"
            newDef = newDef + "\nresult = _df[" + str(params["columns"]) + "]"
            model.getNode(nodeId).definition = self.formatDefinition(newDef)
            return "ok"
        return ""



from pyplan_core.classes.Model import Model
import pandas as pd
from multiprocessing import Lock
import os

os.environ["ENGINE_MODE"]="local"

class Pyplan(object):
    """Main class to interact with the pyplan model"""


    @property
    def sample_models(self):
        return PyplanSampleModels


    def __init__(self):
        self.model = Model()
        self.lastUpdated = -1
        self.currentFilename = ""
        self.lock = Lock()

    def openModel(self, filename):
        """Open Pyplan model"""
        try:
            self._lock_acquire()
            self.currentFilename = filename
            self.lastUpdated = os.path.getmtime(filename)
            self.model.openModel(filename)
        finally:
            self._lock_release()

    def closeModel(self):
        """Close current model"""
        try:
            self._lock_acquire()
            self.model.closeModel()
        finally:
            self._lock_release()

    def getResult(self, node_id):
        """Return result of the node"""
        try:
            self._lock_acquire()
            if self.model.existNode(node_id):
                return self.model.getNode(node_id).result
            raise ValueError(f"The node '{node_id}' was not found in the model")
        finally:
            self._lock_release()

    def setSelectorValue(self, node_id, value):
        """Set value of a selector node"""
        try:
            self._lock_acquire()
            if self.model.existNode(node_id):
                self.model.setSelectorValue(node_id, value)
            else:
                raise ValueError(f"The node '{node_id}' was not found in the model")
        finally:
            self._lock_release()

    def checkForReloadModel(self):
        """Check and reolad model if the last update time has changed"""
        if os.path.getmtime(self.currentFilename) != self.lastUpdated:
            self.openModel(self.currentFilename)

    def getNodeList(self,module_id=None):
        """Return dataframe with the nodes in Pyplan model"""
        arr = []
        if not module_id:
            module_id = self.model.modelNode.identifier
        for node in self.model.findNodes("moduleId",module_id):
            if not node.nodeClass in ["text","alias","inputnode"] and not node.identifier in ["pyplan_library"]: 
                arr.append([node.identifier, node.title, node.nodeClass, node.moduleId])
            
        df = pd.DataFrame(arr, columns=["node_id","title","class","module_id"])
        return df

    def getNode(self, node_id):
        """Return Node from Pyplan model"""
        try:
            self._lock_acquire()
            if self.model.existNode(node_id):
                return self.model.getNode(node_id)
            raise ValueError(f"The node '{node_id}' was not found in the model")
        finally:
            self._lock_release()

    def _lock_acquire(self):
        if not self.lock is None:
            return self.lock.acquire()
        return False

    def _lock_release(self):
        if not self.lock is None:
            try:
                self.lock.release()
            except:
                pass

class PyplanSampleModels(object):
    """Class for read internal sample models"""

    @staticmethod
    def use_of_pyplan_core():
        return os.path.dirname(os.path.abspath(__file__)) + "/sample_models/Sample for Pyplan Core.ppl"



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import json


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "CreateIndex"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            base_node = model.getNode(nodeId)
            new_def = None

            if params["type"]=="list":
                index_values = params["values"]
                new_def = f"result = pd.Index({index_values})"
            elif params["type"]=="range":
                if params["rangeType"]=="string":
                    prefix = params["stringPrefix"]
                    index_values = [f'{prefix}{nn}' for nn in range(int(params["from"]), int(params["to"])+1, int(params["step"]))]
                    str_values = json.dumps(index_values)
                    new_def = f"result = pd.Index({str_values})"
                elif params["rangeType"]=="numeric":
                    index_values = list(range(int(params["from"]), int(params["to"])+1, int(params["step"])))
                    str_values = json.dumps(index_values)
                    new_def = f"result = pd.Index({str_values})"
                elif params["rangeType"]=="date":
                    freq = params["freq"]
                    start = params["start"] if "start" in params and params["start"] else None
                    end = params["end"] if "end" in params and params["end"] else None
                    periods = params["periods"] if "periods" in params and params["periods"] else None
                    if start and end:
                        new_def = f"result = pd.date_range(start='{start}', end='{end}',freq='{freq}')"
                    elif end:
                        new_def = f"result = pd.date_range(end='{end}', periods={periods},freq='{freq}')"
                    else:
                        new_def = f"result = pd.date_range(start='{start}', periods={periods},freq='{freq}')"

            #test and set new definition
            if new_def:
                temp_res = model.evaluate(new_def)
                base_node.definition = new_def
            
        return nodeId



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import jsonpickle


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "SelectRows"

    def getValues(self, model, params):
        field = params["field"]
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            nodeResult = model.getNode(nodeId).result
            if field["type"] == "index":
                pos = nodeResult.index.names.index(field["field"])
                return list(nodeResult.index.levels[pos].values)[:500]
            else:
                return list(nodeResult[field["field"]].unique())[:500]

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            nodeResult = model.getNode(nodeId).result
            currentDef = model.getNode(nodeId).definition
            newDef = currentDef
            conditions = ""
            for nn,filterItem in enumerate(params["filters"]):
                if nn>0:
                    operator = "&"
                    if "logical_operator" in filterItem and filterItem["logical_operator"]:
                        operator = filterItem["logical_operator"]
                    conditions += f" {operator} "

                conditions += self.generateFilter(nodeResult, filterItem)           

            if conditions != "":
                newDef = self.getLastDefinition(currentDef)
                newDef = newDef + "\n" + "# applied filters"
                newDef = newDef + "\n" + "_conditions = " + conditions
                newDef = newDef + "\n" + "result = _df[_conditions]"

            model.getNode(nodeId).definition = self.formatDefinition(newDef)
            return "ok"
        return ""

    def generateFilter(self, nodeResult, filterItem):

        res = ""
        field = filterItem["field"]
        values = filterItem["values"]
        operator = filterItem["operator"]

        selector = ""
        if filterItem["type"] == "column":
            selector = "_df['" + field.replace("'", "\\'") + "']"
        else:
            selector = "_df.index.get_level_values('" + \
                field.replace("'", "\\'") + "')"

        if filterItem["dtype"] == "numeric":
            if operator == "between":
                res = "(" + selector + " > " + str(values[0]) + ")"
                res = res + " & (" + selector + " < " + str(values[1]) + ")"
            elif operator == "outside":
                res = "(" + selector + " < " + str(values[0]) + ")"
                res = res + " & (" + selector + " > " + str(values[1]) + ")"
            elif operator == "defined":
                res = "(~" + selector + ".isnull())"
            elif operator == "undefined":
                res = "(" + selector + ".isnull())"
            else:
                res = "(" + selector + " " + operator + \
                    " " + str(values[0]) + ")"
        else:
            if operator == "contains":
                res = "(" + selector + ".str.contains('" + \
                    str(values[0]).replace("'", "\\'") + "'))"
            elif operator == "notcontains":
                res = "(~" + selector + ".str.contains('" + \
                    str(values[0]).replace("'", "\\'") + "'))"
            else:
                res = "(" + selector + " " + operator + " '" + \
                    str(values[0]).replace("'", "\\'") + "')"
            pass

        return res



from pyplan_core.classes.dynamics.CubepyDynamic import CubepyDynamic
from pyplan_core.classes.dynamics.PureXArrayDynamic import PureXArrayDynamic


class FactoryDynamic(object):

    @staticmethod
    def createInstance(circularNodes, node):
        if not node is None:
            DynamicClass = FactoryDynamic.findDynamicClass(circularNodes, node)
            return DynamicClass()

    @staticmethod
    def findDynamicClass(circularNodes, node):
        for nodeId in circularNodes:
            if node.model.existNode(nodeId):
                _def = node.model.getNode(nodeId).definition
                if "cp.dynamic(" in _def:
                    return CubepyDynamic
                elif "dynamic(" in _def:
                    return PureXArrayDynamic

        return PureXArrayDynamic



import jsonpickle


class IOModule(object):

    def __init__(self, model):
        self.model = model

    def exportModule(self, moduleId, filename):
        res = False
        if self.model.existNode(moduleId):
            toSave = {
                "modelProp": self.model.modelProp,
                "nodeList": []
            }

            # add module
            mainModule = self.model.getNode(moduleId).toObj()
            toSave["nodeList"].append(mainModule)

            # add model
            mainModel = self.model.getNode(mainModule["moduleId"])
            modelObj = mainModel.toObj()
            modelObj["nodeClass"] = "model"
            modelObj["moduleId"] = "_model_"
            toSave["nodeList"].append(modelObj)

            def iterateNodes(subModuleId, toSave):
                for node in self.model.findNodes("moduleId", subModuleId):
                    if(not node.system):
                        toSave["nodeList"].append(node.toObj())
                        if node.nodeClass == "module":
                            iterateNodes(node.identifier, toSave)

            iterateNodes(moduleId, toSave)
            # add childrens

            with open(filename, 'w') as f:
                f.write(jsonpickle.encode(toSave))
                f.close()

            toSave = None
            res = True
        return res

    def importModule(self, moduleId, filename, importType):
        if importType == "0":  # merge
            return self.mergeModule(moduleId, filename)
        elif importType == "2":  # switch
            return self.switchModule(moduleId, filename)
        else:
            return False

    def mergeModule(self, moduleId, filename):
        res = False
        if self.model.existNode(moduleId):
            opened = {}
            with open(filename, 'r') as f:
                opened = jsonpickle.decode(f.read())
                f.close()
            pass
            self.model._isLoadingModel = True

            try:
                for nn, obj in enumerate(opened["nodeList"]):

                    if obj["nodeClass"] != "model":

                        if self.model.existNode(obj["identifier"]):
                            node = self.model.getNode(obj["identifier"])
                            node.definition = obj["definition"]
                            node.title = obj["title"]
                            node.description = obj["description"]
                            node.units = obj["units"]
                            node.color = obj["color"]
                            node.x = obj["x"]
                            node.y = obj["y"]
                            node.w = obj["w"]
                            node.h = obj["h"]
                            node = None
                        else:
                            if nn == 0:
                                obj["moduleId"] = moduleId
                            node = self.model.createNode(
                                obj["identifier"], moduleId=obj["moduleId"])
                            node.fromObj(obj)
                            node = None

                # generate io's
                [self.model.nodeDic[nod].generateIO()
                 for nod in self.model.nodeDic]
            finally:
                self.model._isLoadingModel = False

            res = True

        else:
            raise ValueError('Module base not found')

        return res

    def switchModule(self, moduleId, filename):
        res = False
        if self.model.existNode(moduleId):
            opened = {}
            with open(filename, 'r') as f:
                opened = jsonpickle.decode(f.read())
                f.close()
            pass

            if len(opened["nodeList"]) > 0:
                mainModule = opened["nodeList"][0]
                mainId = mainModule["identifier"]
                self.model.deleteNodes([mainId], removeAliasIfNotIn=mainId)

            self.model._isLoadingModel = True

            try:
                for nn, obj in enumerate(opened["nodeList"]):
                    if nn == 0:
                        obj["moduleId"] = moduleId

                    if obj["nodeClass"] != "model":

                        if self.model.existNode(obj["identifier"]):
                            node = self.model.getNode(obj["identifier"])
                            node.identifier = node.identifier + "_copy"

                        node = self.model.createNode(
                            obj["identifier"], moduleId=obj["moduleId"])
                        node.fromObj(obj)
                        node = None

                # generate io's
                for nodeId in self.model.nodeDic:
                    self.model.nodeDic[nodeId].generateIO()
            finally:
                self.model._isLoadingModel = False

            res = True

        else:
            raise ValueError('Module base not found')

        return res



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import xarray as xr
import json


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "DataArrayFilter"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId) and "filters" in params:
            nodeResult = model.getNode(nodeId).result
            current_def = model.getNode(nodeId).definition
            new_def = current_def
            filter_dict = params["filters"]

            new_def = self.getLastDefinition(current_def, "_da")
            new_def = f"{new_def}\n# applied filters"
            new_def = f"{new_def}\nresult = _da.sel({json.dumps(filter_dict)}, drop=True)"

            model.getNode(nodeId).definition = self.formatDefinition(new_def)
            return nodeId
        return ""



from enum import Enum


class filterChoices(Enum):
    CONTAINS = 'contains'
    NOT_CONTAINS = 'notcontains'
    EQUAL = 'equal'
    NOT_EQUAL = 'notequal'
    GREATER_THAN = 'greater_than'
    GREATER_THAN_EQUAL_TO = 'greater_than_equal_to'
    LESS_THAN = 'less_than'
    LESS_THAN_EQUAL_TO = 'less_than_equal_to'
    BEGIN_WITH = 'begin_with'
    NOT_BEGIN_WITH = 'not_begin_with'
    END_WITH = 'end_with'
    NOT_END_WITH = 'not_end_with'
    EARLIEST = 'earliest'
    LATEST = 'latest'
    BETWEEN = 'between'
    NOT_BETWEEN = 'notbetween'

    def __str__(self):
        return self.value



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import jsonpickle


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "sourcecsv"

    def getColumns(self, model, params):
        res = []
        current_path = model.getNode("current_path").result
        params["rows"] = 5000
        _df = self._openFile(params, current_path)
        totalReg = _df.shape[0]

        # append columns
        for nn, cols in enumerate(list(_df.columns)):
            if not cols is None:
                item = dict(field=cols, type="column",
                            dtype=self.kindToString(_df.dtypes[nn].kind))
                if (item["dtype"] != "numeric"):
                    uniqueCount = len(_df[cols].unique())
                    if uniqueCount/totalReg < 0.8:
                        item["type"] = "index"
                res.append(item)

        return res

    def readFile(self, model, params):
        current_path = model.getNode("current_path").result
        _df = self._openFile(params, current_path)
        res = {
            "params": params,
            "result": jsonpickle.decode(_df.to_json(orient="table"))
        }
        return res

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]

        charcode = (None if params["encoding"] ==
                    "auto" else params["encoding"])
        sep = (None if params["sep"] == "auto" else params["sep"])
        index_col = None
        if "indexes" in params and len(params["indexes"]) > 0:
            index_col = params["indexes"]

        newDef = "result = pd.read_csv(current_path + '{}'".format(
            params["filename"])
        if params["filename"].lower().startswith("http"):
            newDef = "result = pd.read_csv('{}'".format(params["filename"])

        if not charcode is None:
            newDef = newDef + ", encoding='{}'".format(charcode)
        if sep is None:
            newDef = newDef + ", sep=None"
        else:
            newDef = newDef + ", sep='{}'".format(sep)
        if not index_col is None:
            newDef = newDef + ", index_col={}".format(index_col)
        newDef = newDef + ")"

        model.getNode(nodeId).definition = self.formatDefinition(newDef)
        return "ok"

    def _openFile(self, params, current_path):
        filename = current_path + params["filename"]
        if params["filename"].lower().startswith("http"):
            filename = params["filename"]

        sample = 5000
        charcode = (None if params["encoding"] ==
                    "auto" else params["encoding"])
        sep = (None if params["sep"] == "auto" else params["sep"])

        index_col = None
        if "indexes" in params and len(params["indexes"]) > 0:
            index_col = params["indexes"]

        codecs = [charcode, "utf-8", "cp850", "cp852",
                  "cp1250", "cp1252", "latin_1", "ascii"]
        nChance = 0

        while nChance < len(codecs):
            try:
                tryCharcode = codecs[nChance]
                _df = pd.read_csv(filename, encoding=tryCharcode,
                                  sep=sep, nrows=sample, index_col=index_col)

                if charcode is None and not tryCharcode is None:
                    params["encoding"] = tryCharcode

                return _df.head(int(params["rows"]))

            except Exception as ex:
                if "codec can't decode byte" in str(ex) and nChance+1 < len(codecs):
                    nChance += 1
                else:
                    raise ex



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import jsonpickle
import re


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "CalculatedField"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        formula = params["formula"]
        name = params["name"]
        if model.existNode(nodeId):
            currentDef = model.getNode(nodeId).definition
            newDef = self.getLastDefinition(currentDef)
            newDef = newDef + "\n# Generated new column"

            # find fields on formula
            regex = r"#(.+?)#"
            matches = re.finditer(regex, formula)
            formulaFields = []
            for matchNum, match in enumerate(matches):
                formulaFields.append(str(match.group(1)))

            # find df fields
            dfFields = self.getColumnList(model, params)

            # replace def
            for field in formulaFields:
                dfField = [f for f in dfFields if f["field"] == field]
                if not dfField is None and len(dfField) == 1:
                    fieldDef = ""
                    if dfField[0]["type"] == "column":
                        fieldDef = "_df['" + field.replace("'", "\\'") + "']"
                    else:
                        fieldDef = "_df.index.get_level_values('" + field.replace(
                            "'", "\\'") + "')"

                    key = "#"+field+"#"
                    formula = formula.replace(key, fieldDef)

            newDef = newDef + "\n_df['" + \
                name.replace("'", "") + "'] = " + formula
            newDef = newDef + "\nresult = _df"

            model.getNode(nodeId).definition = self.formatDefinition(newDef)
            return "ok"
        return ""



import json

import numpy as np
import pandas as pd
import xarray as xr

from pyplan_core import cubepy 
from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator
from pyplan_core.classes.common.filterChoices import filterChoices
from pyplan_core.classes.common.indexValuesReq import IndexValuesReq
from pyplan_core.cubepy.cube import kindToString, safemax, safemean, safemin, safesum


class CubepyEvaluator(BaseEvaluator):

    PAGESIZE = 100
    MAX_COLUMS = 5000

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        if isinstance(result, cubepy.Cube):
            return self.cubeEvaluate(result, nodeDic, nodeId, dims, rows, columns, summaryBy, bottomTotal, rightTotal, fromRow, toRow)
        elif isinstance(result, cubepy.Index):
            return self.indexEvaluate(result, nodeDic, nodeId, dims, rows, columns, summaryBy, bottomTotal, rightTotal, fromRow, toRow)

    def cubeEvaluate(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        result_structure = self.getStructure(result)
        sby = safesum
        if summaryBy == 'avg':
            sby = safemean
        elif summaryBy == 'max':
            sby = safemax
        elif summaryBy == 'min':
            sby = safemin

        if (fromRow is None) or int(fromRow) <= 0:
            fromRow = 1
        if (toRow is None) or int(toRow) < 1:
            toRow = 100
        fromRow = int(fromRow)
        toRow = int(toRow)

        result = self.applyHierarchy(
            result, nodeDic, nodeId, dims, rows, columns, sby)

        _filters = []
        _rows = []
        _columns = []
        if not rows is None:
            for row in rows:
                if self.hasDim(result, str(row["field"])):
                    _rows.append(str(row["field"]))
                    self.addToFilter(nodeDic, row, _filters)

        if not columns is None:
            for column in columns:
                if self.hasDim(result, str(column["field"])):
                    _columns.append(str(column["field"]))
                    self.addToFilter(nodeDic, column, _filters)

        if not dims is None:
            for dim in dims:
                if self.hasDim(result, str(dim["field"])):
                    self.addToFilter(nodeDic, dim, _filters)

        tmp = None
        if len(_rows) == 0 and len(_columns) == 0 and result.ndim > 0:
            #_rows.append( result.dims[0] )
            tmp = cubepy.Cube([], result.filter(_filters).reduce(sby))

        else:
            tmp = result.filter(_filters).reduce(sby, keep=(
                _rows + _columns)).transpose(_rows + _columns)

        finalValues = tmp.values
        finalIndexes = []
        if tmp.ndim > 0:
            finalIndexes = tmp.axes[0].values
        finalColumns = ["Total"]
        if tmp.ndim == 2:
            finalColumns = tmp.axes[1].values

        # Add totales
        _totalRow = None
        if bottomTotal and len(_rows) > 0:
            # add total row
            #finalIndexes = np.append(finalIndexes,"Total")
            if tmp.ndim == 1:
                _totalRow = finalValues.sum(axis=0).reshape(1)
                #finalValues = np.append( finalValues, finalValues.sum(axis=0).reshape(1), axis=0)
            else:
                _totalRow = finalValues.sum(
                    axis=0).reshape(1, len(finalValues[0]))
                _totalRow = _totalRow[0]
                if rightTotal:
                    _totalRow = np.append(_totalRow, finalValues.sum())

        if rightTotal and len(_columns) > 0:
            # add total column
            if tmp.ndim == 1:
                finalIndexes = np.append(finalIndexes, "Total")
                finalValues = np.append(
                    finalValues, finalValues.sum(axis=0).reshape(1), axis=0)
            else:
                finalColumns = np.append(finalColumns, "Total")
                finalValues = np.append(finalValues, finalValues.sum(
                    axis=1).reshape(len(finalValues), 1), axis=1)

            # con una sola dimension

        # chek inf
        if kindToString(finalValues.dtype.kind) == "numeric":
            if np.isinf(finalValues).any():
                finalValues[np.isinf(finalValues)] = None

        # chec if haver nan values
        # if np.isnan(finalValues).any():
        if pd.isnull(finalValues).any():
            try:
                finalValues = np.where(
                    np.isnan(finalValues), None, finalValues)
            except:
                finalValues[pd.isnull(finalValues)] = None

        res = {}
        pageInfo = None
        onRow = None
        onColumn = None
        if len(_rows) == 0 and len(_columns) == 0:
            res = {
                "columns": [],
                "index": ["Total"],
                "data": [[finalValues.tolist()]]
            }
        elif len(_rows) == 0:
            onColumn = _columns[0]
            res = {
                "columns": finalIndexes[:CubepyEvaluator.MAX_COLUMS].tolist(),
                "index": finalColumns,
                "data": [finalValues[:CubepyEvaluator.MAX_COLUMS].tolist()]
            }
        elif len(_columns) == 0:

            if (len(finalIndexes) > self.PAGESIZE):
                pageInfo = {
                    "fromRow": int(fromRow),
                    "toRow": int(toRow),
                    "totalRows": len(finalIndexes)
                }

            onRow = _rows[0]
            res = {
                "columns": finalColumns,
                "index": finalIndexes[fromRow-1:toRow].tolist(),
                "data": [[x] for x in finalValues[fromRow-1:toRow].tolist()]
            }
            # add total rows
            if not _totalRow is None:
                res["index"].append("Total")
                res["data"].append(_totalRow.tolist())

        else:
            onColumn = _columns[0]
            onRow = _rows[0]

            if (len(finalIndexes) > self.PAGESIZE):
                pageInfo = {
                    "fromRow": int(fromRow),
                    "toRow": int(toRow),
                    "totalRows": len(finalIndexes)
                }

            res = {
                "columns": finalColumns[:CubepyEvaluator.MAX_COLUMS].tolist(),
                "index": finalIndexes[fromRow-1:toRow].tolist(),
                "data": finalValues[fromRow-1:toRow, :CubepyEvaluator.MAX_COLUMS].tolist()
            }

            # add total rows
            if not _totalRow is None:
                res["index"].append("Total")
                res["data"].append(
                    _totalRow[:CubepyEvaluator.MAX_COLUMS].tolist())

        return self.createResult(res, result_structure, onRow=onRow, onColumn=onColumn, node=nodeDic[nodeId], pageInfo=pageInfo)

    def applyHierarchy(self, result, nodeDic, nodeId, dims, rows, columns, sby):

        def hierarchize(cube, levels, maps):

            mapArray = nodeDic[maps[0]].result
            coordValues = mapArray.values.copy()
            targetIndexId = nodeDic[levels[1]].result.name

            for pos, level in enumerate(levels):
                if pos > 0:

                    if not maps[pos] is None:
                        mapArrayLevel = nodeDic[maps[pos]].result

                        for ii in range(len(coordValues)):
                            if not coordValues[ii] is None:
                                try:
                                    newVal = mapArrayLevel.filter(
                                        mapArrayLevel.dims[0], coordValues[ii]).values.item(0)
                                    coordValues[ii] = newVal
                                except Exception as ex:
                                    coordValues[ii] = None
                                    #raise ValueError("Hierarchy not found. Level: " + targetIndexId + ", value: " + coordValues[ii])
                                    pass

            # convert to dataarray
            _coords = []
            for _axis in cube.axes:
                _coord = pd.Index(_axis.values, name=_axis.name)
                _coords.append(_coord)
            dataArray = xr.DataArray(cube.values, _coords)

            # perform aggregate
            dataArray.coords[levels[0]].values = coordValues
            _df = dataArray.to_series()
            _df = _df.groupby(list(dataArray.dims), sort=False).agg(sby)
            _da = _df.to_xarray()
            reindex_dic = dict()
            for dimension in _da.dims:
                if dimension == levels[0]:
                    reindex_dic[dimension] = nodeDic[levels[-1:]
                                                     [0]].result.values
                elif dimension in nodeDic and isinstance(nodeDic[dimension].result, cubepy.Index):
                    reindex_dic[dimension] = nodeDic[dimension].result.values

            _db = _da.reindex(reindex_dic)

            # convert to cube
            _indexes = []
            for _dim in _db.dims:
                _index = cubepy.Index(_dim, _db.coords[_dim].values)
                _indexes.append(_index)
            _cube = cubepy.Cube(_indexes, _db.values)

            return _cube

        allDims = (dims or []) + (rows or []) + (columns or [])

        for dim in allDims:
            if dim and dim["currentLevel"] and dim["currentLevel"] != str(dim["field"]).split(".")[0]:
                # Find the path for the current level

                # recursive fn for search parent
                def findPath(indexNode, level, levels, maps):
                    if indexNode.identifier == level:
                        levels.append(indexNode.identifier)
                        maps.append(None)
                        return True
                    else:
                        parents = indexNode.hierarchy_parents
                        if parents is None:
                            return False
                        else:
                            if not isinstance(parents, list):
                                parents = [parents]
                            mapArrays = indexNode.hierarchy_maps
                            if not isinstance(mapArrays, list):
                                mapArrays = [mapArrays]

                            mapPos = 0
                            for parentId in parents:
                                parent = nodeDic[parentId]
                                if findPath(parent, level, levels, maps):
                                    levels.append(indexNode.identifier)
                                    maps.append(mapArrays[mapPos])
                                    return True
                                mapPos += 1
                    return False

                field = str(dim["field"]).split(".")[0]
                currentLevel = dim["currentLevel"]
                indexNode = nodeDic[field]

                levels = []
                maps = []
                findPath(indexNode, currentLevel, levels, maps)
                levels.reverse()
                maps.reverse()

                result = hierarchize(cubepy.cube.Cube(
                    result._axes, result._values), levels, maps)
        return result

    def indexEvaluate(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        res = result.values[:100].tolist()
        return self.createResult(res, self.getStructure(res), node=nodeDic[nodeId])

    def addToFilter(self, nodeDic, dim, filters):
        if "values" in dim and dim["values"] is not None and len(dim["values"]) > 0:
            #field = str(dim["field"]).split(".")[0]
            field = str(dim["field"])
            nodeId = None
            # if len(str(dim["field"]).split("."))>1:
            #     nodeId = str(dim["field"]).split(".")[1]
            indexType = None
            # is from numpy evaluator
            if "." in field and field.startswith("Axis"):
                indexType = "N"
            else:
                indexType = self.getIndexType(nodeDic, nodeId, field)

            if indexType == "S":
                filters.append(cubepy.Index(
                    field,  [str(v["value"]) for v in dim["values"]]))
            else:
                filters.append(cubepy.Index(
                    field,  [int(v["value"]) for v in dim["values"]]))

    def getCubeValues(self, result, nodeDic, nodeId, query):
        if isinstance(result, cubepy.Cube):
            res = {
                "dims": [],
                "values": []
            }

            _filters = []
            if not query["filters"] is None:
                for dimFilter in query["filters"]:
                    #field = str(dimFilter["field"]).split(".")[0]
                    field = str(dimFilter["field"])
                    if self.hasDim(result, field):
                        indexType = None
                        # is from numpy evaluator
                        if "." in field and field.startswith("Axis"):
                            indexType = "N"
                        else:
                            indexType = self.getIndexType(
                                nodeDic, nodeId, field)

                        if indexType == "S":
                            _filters.append(cubepy.Index(
                                field,  [str(v) for v in dimFilter["values"]]))
                        else:
                            _filters.append(cubepy.Index(
                                field,  [int(v) for v in dimFilter["values"]]))

            _filteredResult = result.filter(_filters)
            for col in query["columns"]:
                if col in self.getIndexes(nodeDic[nodeId], result):
                    item = {
                        "field": col,
                        "count": 0,
                        "values": [str(v) for v in _filteredResult.axis(col).values.tolist()]
                        # "values": result.filter(_filters).axis(col).values.tolist()
                    }
                    item["count"] = len(item["values"])
                    res["dims"].append(item)

            resultValues = _filteredResult.sum(keep=query["columns"])
            if isinstance(resultValues, cubepy.Cube):
                res["values"] = resultValues.transpose(
                    query["columns"]).values.reshape(resultValues.size).tolist()
            elif isinstance(resultValues, str):
                res["values"] = [resultValues]
            else:
                res["values"] = resultValues
            return res

    def getCubeDimensionValues(self, result, nodeDic, nodeId, query):
        if isinstance(result, cubepy.Cube):
            if len(query["columns"]) > 0:
                dimension = query["columns"][-1]

            if dimension in self.getIndexes(nodeDic[nodeId], result):
                finalList = [str(v) for v in result.axis(
                    dimension).values.tolist()[:1000]]
                finalList.sort()
                return finalList
        return []

    def getCubeMetadata(self, result, nodeDic, nodeId):
        res = None
        if isinstance(result, cubepy.Cube):
            res = {
                "dims": [],
                "measures": [],
                "aggregator": "sum",
                "isEditable": True if self.isTable(nodeDic[nodeId]) == "1" else False,
                "nodeProperties": {
                    "title": nodeDic[nodeId].title if not nodeDic[nodeId].title is None else nodeDic[nodeId].identifier,
                    "numberFormat": nodeDic[nodeId].numberFormat
                }
            }

            # check if is in scenario
            if nodeDic[nodeId].model.isNodeInScenario(nodeDic[nodeId].identifier):
                res["nodeProperties"]["scenario"] = True

            for dim in result.dims:
                itemDim = {
                    "field": dim,
                    "name": str(dim).split(".")[0]
                }
                if dim in nodeDic:
                    if not nodeDic[dim].title is None:
                        itemDim["name"] = nodeDic[dim].title
                res["dims"].append(itemDim)

            res["measures"].append({
                "field": "datavalue",
                "name": "datavalue"
            })

        return res

    def setNodeValueChanges(self, nodeDic, nodeId, nodeChanges):
        if isinstance(nodeDic[nodeId].result, cubepy.Cube):
            for change in nodeChanges["changes"]:
                newValue = change["definition"]
                filters = []
                for filterItem in change["filterList"]:
                    aux = {
                        "field": filterItem["Key"],
                        "values": [{
                            "value": filterItem["Value"]
                        }]}
                    self.addToFilter(nodeDic, aux, filters)
                nodeDic[nodeId].result.set_data(filters, newValue)
                # generate and put new definition bypass clear result
                newDeff = nodeDic[nodeId].result.generateDefinition()
                nodeDic[nodeId].definition = newDeff

            return "ok"

    def isTable(self, node):
        res = "0"
        if isinstance(node.result, cubepy.Cube):
            if not node.definition is None and node.definition != "":
                import re
                deff = re.sub(
                    '[\s+]', '', str(node.definition).strip(' \t\n\r')).lower()
                if (deff.startswith("this=cubepy.cube(") or deff.startswith("result=cubepy.cube(")
                        or deff.startswith("this=cp.cube(") or deff.startswith("result=cp.cube(")
                        or deff.startswith("this=cubepy.cube.zeros(") or deff.startswith("result=cubepy.cube.zeros(")
                        or deff.startswith("this=cubepy.cube.ones(") or deff.startswith("result=cubepy.cube.ones(")
                        or deff.startswith("this=cubepy.cube.full(") or deff.startswith("result=cubepy.cube.full(")):
                    res = "1"
        elif isinstance(node.result, np.ndarray):
            if not node.definition is None and node.definition != "":
                import re
                deff = re.sub(
                    '[\s+]', '', str(node.definition).strip(' \t\n\r')).lower()
                if (deff.startswith("result=np.")):
                    res = "1"

        return res

    def hasDim(self, result, dim):
        return True if dim in result.dims else False

    def getIndexes(self, node, result=None):
        res = []
        if result is None:
            result = node._result

        if isinstance(result, cubepy.Cube):
            res = list(result.dims)
        return res

    def getIndexesWithLevels(self, node, result=None):
        res = []
        if result is None:
            result = node._result

        if isinstance(result, cubepy.Cube):
            if not result is None:
                _model = node.model
                for indexItem in result.dims:
                    itemDim = indexItem
                    item = {"field": itemDim, "name": itemDim,
                            "description": "", "levels": []}

                    if node.model.existNode(itemDim):
                        levelNode = node.model.getNode(itemDim)
                        if levelNode.title:
                            item["name"] = levelNode.title
                            item["description"] = levelNode.description

                        # check for levels
                        if not levelNode.hierarchy_parents is None:

                            def buildLevels(parents, levelList):
                                if not isinstance(parents, list):
                                    parents = [parents]
                                for parentIndexId in parents:
                                    parentIndexNode = _model.getNode(
                                        parentIndexId)
                                    if parentIndexNode is None:
                                        raise ValueError(
                                            f"Node {parentIndexId} not found")

                                    levelItem = {
                                        "field": parentIndexId, "name": parentIndexNode.title or parentIndexId}
                                    levelList.append(levelItem)
                                    _dummy = parentIndexNode.result  # to force calc
                                    if not parentIndexNode.hierarchy_parents is None:
                                        buildLevels(
                                            parentIndexNode.hierarchy_parents, levelList)

                            listOfLevels = [
                                {"field": itemDim, "name": item["name"]}]
                            indexParents = levelNode.hierarchy_parents
                            buildLevels(indexParents, listOfLevels)
                            item["levels"] = listOfLevels
                    res.append(item)

        return res

    def getIndexType(self, nodeDic, nodeId, indexId):
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        res = "S"
        if (not indexId is None) & (indexId in nodeDic):
            node = nodeDic[indexId]
            if isinstance(node.result, cubepy.Index):
                if str(node.result.values.dtype) in numerics:
                    res = "N"
                else:
                    res = "S"
            elif isinstance(node.result, np.ndarray):
                if str(node.result.dtype) in numerics:
                    res = "N"
                else:
                    res = "S"
            else:
                res = "S"
        return res

    def getIndexValues(self, nodeDic, data: IndexValuesReq, result=None):
        res = []
        if (not data.index_id is None) & (data.index_id in nodeDic):
            node = nodeDic[data.index_id]
            if isinstance(node.result, cubepy.Index):
                res = (node.result.values).tolist()
            elif isinstance(node.result, np.ndarray):
                res = (node.result).tolist()
            else:
                res = list(node.result)
        if data.text1:
            text1 = data.text1.lower()
            if data.filter == filterChoices.CONTAINS.value:
                res = list(
                    filter(lambda item: text1 in str(item).lower(), res))
            elif data.filter == filterChoices.NOT_CONTAINS.value:
                res = list(
                    filter(lambda item: not text1 in str(item).lower(), res))
        return res

    def previewNode(self, nodeDic, nodeId):
        from pyplan_core.cubepy.Helpers import Helpers
        from sys import getsizeof
        res = {
            "resultType": str(type(nodeDic[nodeId].result)),
            "dims": [],
            "console": nodeDic[nodeId].lastEvaluationConsole,
            "preview": ""
        }
        if isinstance(nodeDic[nodeId].result, cubepy.Cube):
            cube = nodeDic[nodeId].result
            for _axis in cube.axes:
                _nodeTitle = None
                if _axis.name in nodeDic:
                    _nodeTitle = nodeDic[_axis.name].title

                if _nodeTitle is None:
                    _item = _axis.name + " [" + str(len(_axis)) + "]"
                else:
                    _item = _nodeTitle + \
                        " ("+_axis.name+") [" + str(len(_axis)) + "]"
                res["dims"].append(_item)

            res["preview"] += "Dimensions: " + str(cube.ndim)
            res["preview"] += "\nShape: " + str(cube.shape)
            res["preview"] += "\nSize: " + str(cube.size)
            res["preview"] += "\nMemory: " + \
                str(round(getsizeof(cube)/1024/1024, 2)) + " Mb"
            if not cube.values is None:
                res["preview"] += "\nData type: " + \
                    str(cube.values.dtype) + \
                    " (" + kindToString(cube.values.dtype.kind)+")"
                res["preview"] += "\nValues: \n\n" + str(cube.values)[:1000]

        elif isinstance(nodeDic[nodeId].result, cubepy.Index):
            index = nodeDic[nodeId].result

            res["preview"] += "Size: " + str(len(index))
            res["preview"] += "\nMemory: " + \
                str(round(getsizeof(index)/1024/1024, 2)) + " Mb"
            if not index.values is None:
                res["preview"] += "\nData type: " + \
                    str(index.values.dtype) + \
                    " (" + kindToString(index.values.dtype.kind)+")"
                #res["preview"] += "\nValues: \n" + '\n\t'.join([''.join(row) for row in index.values[:100]])
                res["preview"] += "\nValues: \n\t" + \
                    '\n\t'.join([''.join(str(row))
                                 for row in index.values[:100]])

        return json.dumps(res)

    def dumpNodeToFile(self, nodeDic, nodeId, fileName):
        definition = nodeDic[nodeId].result.generateDefinition()
        with open(fileName, 'w') as f:
            f.write(definition)
            f.close()

    def geoUnclusterData(self, result, nodeDic, nodeId, rowIndex, attIndex, latField="latitude", lngField="longitude", geoField="geoField", labelField="labelField", sizeField="sizeField", colorField="colorField", iconField="iconField"):
        latField = "latitude" if latField is None else latField
        lngField = "longitude" if lngField is None else lngField
        geoField = "geoField" if geoField is None else geoField
        labelField = "labelField" if labelField is None else labelField
        sizeField = "sizeField" if sizeField is None else sizeField
        colorField = "colorField" if colorField is None else colorField
        iconField = "iconField" if iconField is None else iconField

        _tmp_for_geo = cubepy.Index('tmp_for_geo', [
                                    latField, lngField, geoField, labelField, sizeField, colorField, iconField])
        _idx = nodeDic[attIndex].result
        rowIndexObj = nodeDic[rowIndex].result
        mapCube = result[_idx == _tmp_for_geo].transpose(
            [rowIndex, "tmp_for_geo"]).values
        res = dict()
        points = []
        pos = 0
        for itemRow in mapCube:
            vo = dict()
            vo["id"] = str(rowIndexObj.values[pos])
            vo["lat"] = itemRow[0]
            vo["lng"] = itemRow[1]
            vo["geoDef"] = itemRow[2]
            vo["labelRes"] = itemRow[3]
            vo["sizeRes"] = itemRow[4]
            vo["colorRes"] = itemRow[5]
            vo["iconRes"] = itemRow[6]
            points.append(vo)
            pos += 1
        res["points"] = points

        return res



from setuptools import setup
import os

REQ_FILE = os.path.join(os.path.dirname(__file__), 'requirements.txt')

required = []
with open(REQ_FILE) as f:
    required = f.read().splitlines()

setup(install_requires=required)



"""Utility functions imported by cubepy.cube and cubepy.axes.
Do not import these in order to avoid circular imports.
"""

from pyplan_core import cubepy 
import pyplan_core.cubepy.axis
import pyplan_core.cubepy.index
from pyplan_core.cubepy.exceptions import AxisAlignError

import numpy as np


def is_axis(obj):
    return isinstance(obj, cubepy.axis.Axis)


def is_indexed(axis):
    return hasattr(axis, "indexof")


def make_axis_collection(axes):
    """Creates a list of axes if a single axis is passed in."""
    if isinstance(axes, int) or isinstance(axes, str) or is_axis(axes):
        return [axes]
    else:
        return axes


def align_arrays(axis1, axis2, axis_index1, axis_index2, values1, values2):
    """
    :param axis1:
    :param axis2:
    :param axis_index1:
    :param axis_index2:
    :param values1:
    :param values2:
    :return: tuple (axis, values1, values2)
    """
    if axis1 is axis2:
        # if self alignment, then do nothing
        return axis1, None, None
    elif is_indexed(axis2):
        # align second axis to first axis
        value_indices = axis2.indexof(axis1.values)
        return axis1, values1, values2.take(value_indices, axis_index2)
    elif is_indexed(axis1):
        # align first axis to second axis
        value_indices = axis1.indexof(axis2.values)
        return axis2, values1.take(value_indices, axis_index1), values2
    else:  # both are non-indexed
        if not np.array_equal(axis1.values, axis2.values):
            raise AxisAlignError(
                "cannot align axes '{}' with unequal values".format(axis1.name))
        return axis1, values1, values2


def broadcast_array(values, old_axes, new_axes):
    """Add new virtual axes (length is 1) to a numpy array to correspond to the new axes."""
    new_values = values
    transpose_indices = []
    for axis in new_axes:
        try:
            axis_index = old_axes.index(axis.name)
        except LookupError:
            # if axis is not present in the cube, add virtual axis at the end
            axis_index = new_values.ndim
            new_values = np.expand_dims(new_values, axis=axis_index)
        transpose_indices.append(axis_index)

    # handle the trailing axes
    if new_values.ndim != len(new_axes):
        raise ValueError("cube broadcasting axis mismatch")

    # transpose the result
    return new_values.transpose(transpose_indices)


def unique_axes_from_cubes(cubes):
    """Creates common axis space for a collection of cubes. The following rules are observed:
    1) axes are identified only by their name, values are ignored
    2) axes are listed as they are listed in the cube collection
    3) non-indexable axes have priority over the indexable with the same name
    4) axes are not aligned by values, values are ignored
    """
    unique_axes_list = list()
    unique_axes_dict = dict()

    for cube in cubes:
        for axis in cube.axes:
            try:
                base_axis_index = unique_axes_dict[axis.name]
            except KeyError:
                # add new axis
                unique_axes_dict[axis.name] = len(unique_axes_list)
                unique_axes_list.append(axis)
            else:
                # axis with the same name was found
                base_axis = unique_axes_list[base_axis_index]
                if not is_indexed(axis) and is_indexed(base_axis):
                    # replace indexed axis with fixed axis
                    unique_axes_list[base_axis_index] = axis

    return unique_axes_list



import numpy as np

from pyplan_core.cubepy.exceptions import NonUniqueDimNamesError
from pyplan_core.cubepy.utils import is_axis


class Axes(object):
    """Axes is an internal helper class allowing easier manipulation with cube axes.
    It is not intended to be accessed by users of cubepy package.

    Get axis name of an axis of given index:
    ax_name = cube.axes[0].name

    Get index of axis of a given name:
    ax_index = cube.axes.index('a')

    Axes can be indexed using integer or string index:
    ax1 = cube.axes[0]
    ax_a = cube.axes['a']

    Axes can be used as iterables:
    axes_list = list(cube.axes)  
    # the above is better than [axis for axis in cube.axes]
    axes_dict = dict(name: axis for name, axis in enumerate(cube.axes))"""

    def __init__(self, axes):
        """If for non-unique axes are found, ValueError is raised.
        If axis has invalid type, TypeError is raised.
        :param axes: Axis or a collection of Axis objects (incl. another Axes object)
        """
        # special case with zero axes
        if axes is None:
            self.axes = tuple()
            self.dims = tuple()
            return

        # special case with a single axis
        if is_axis(axes):
            axes = [axes]

        unique_names = set()
        for axis in axes:
            # test correct types
            if not is_axis(axis):
                raise TypeError("axis must be an instance of Axis")
            # test unique names - report the name of the first axis which is not unique
            if axis.name in unique_names:
                raise NonUniqueDimNamesError(
                    "multiple axes with name '{}'".format(axis.name))
            unique_names.add(axis.name)

        # the sequence of axes must be immutable
        self.axes = tuple(axes)
        self.dims = tuple(a.name for a in axes)

    def __repr__(self):
        axes = [str(a) for a in self.axes]
        return "Axes(" + ', '.join(axes) + ")"

    def __len__(self):
        return len(self.axes)

    def __getitem__(self, index):
        """Return an axis given by its index.
        :param index: int
        :return: an Axis object
        :raise: IndexError if not found
        """
        return self.axes[index]

    def axis_by_index(self, index):
        return self.axes[index]

    def axis_by_name(self, name):
        """Returns None if not found.
        """
        return next((axis for axis in self.axes if axis.name == name), None)

    def axis_and_index(self, axis):
        index = self.index(axis)
        return self[index], index

    def is_unique_subset(self, axes):
        """Tests whether all axes are contained in the Axes object and whether they are unique.
        """
        raise NotImplementedError

    def complement(self, axes):
        """Return a tuple of indices of axes from Axes object which are not
        contained in the specified collection of axes.
        :param axes: collection of axes specified by axis name or index
        :return: tuple of int
        """
        if isinstance(axes, str) or isinstance(axes, int):
            axes = (axes,)
        indices = set(self.index(a) for a in axes)
        if len(indices) != len(axes):
            raise ValueError("axes are not unique")
        return tuple(i for i in range(len(self)) if i not in indices)

    def index(self, axis):
        """Find axis index by name, by index, or by axis object.
         :param axis: int, str or Axis
         :return: int
         :raise: LookupError if not found
         """
        # find by numeric index, normalize negative numbers
        if isinstance(axis, int):
            axis_count = len(self.axes)
            if 0 <= axis < axis_count:
                return axis
            if -axis_count <= axis < 0:
                # negative index is counted from the last axis backward
                return axis_count + axis
            raise LookupError("invalid axis index: {}".format(axis))

        # find by name
        if isinstance(axis, str):
            for i, a in enumerate(self.axes):
                if a.name == axis:
                    return i
            raise LookupError("invalid axis name: '{}'".format(axis))

        # find by object identity
        if is_axis(axis):
            for i, a in enumerate(self.axes):
                if a is axis:
                    return i
            # try by name
            return self.index(axis.name)
            #raise LookupError("axis not found: {}".format(axis))

        raise TypeError("invalid axis identification type")

    def contains(self, axis):
        """Returns True/False indicating whether the axis is contained in the Axes object.
        Axis can be specified by name (str), by index (int) or by Axis object.
        """
        try:
            self.index(axis)
            return True
        except LookupError:
            return False

    def transposed_indices(self, front, back):
        """Reorder axes by specified names or indices. Return a list of axis
        indices which correspond to the new order of axes.
        """
        if isinstance(front, str) or isinstance(front, int) or is_axis(front):
            front = [front]

        if isinstance(back, str) or isinstance(back, int) or is_axis(back):
            back = [back]

        front_axes = list()
        back_axes = list()
        temp_axes = list(self.axes)
        for axis_id in front:
            index = self.index(axis_id)
            front_axes.append(index)
            if temp_axes[index] is None:
                raise ValueError("duplicate axes in transpose")
            temp_axes[index] = None

        for axis_id in back:
            index = self.index(axis_id)
            back_axes.append(index)
            if temp_axes[index] is None:
                raise ValueError("duplicate axes in transpose")
            temp_axes[index] = None

        middle_axes = [index for index, axis in enumerate(
            temp_axes) if axis is not None]
        return front_axes + middle_axes + back_axes

    def insert(self, axis, index=0):
        """Insert a new axis at the specified position and return the new Axes object.
        :param axis: the new axis to be inserted
        :param index: the index of the new axis
        :return: new Axes object
        """
        axis_list = list(self.axes)
        # need to correctly handle negative values
        # for example: index -1 means that the new axis should be the last axis after the insertion
        if index < 0:
            index = len(axis_list) + 1 - index
        axis_list.insert(index, axis)
        return Axes(axis_list)

    def remove(self, axis_id):
        """Remove axis or axes with a given index or name.
        Return new Axes object.
        """
        axis_index = self.index(axis_id)
        return self._remove(axis_index)

    def _remove(self, axis_index):
        new_axes = list(self.axes)
        del new_axes[axis_index]
        return Axes(new_axes)

    def rename(self, old_axis_id, new_axis_name):
        """Return an Axes object with a renamed axis.
        :param old_axis_id: axis index (int) or name (str)
        :param new_axis_name: the name of the new axis (str)
        :return: new Axes object
        """
        old_axis, old_axis_index = self.axis_and_index(old_axis_id)
        new_axis = old_axis.rename(new_axis_name)
        return self._replace(old_axis_index, new_axis)

    def replace(self, old_axis_id, new_axis):
        """Replace an existing axis with a new axis and return the new Axes object.
        The new axes collection is checked for duplicate names.
        The old and new axes are NOT checked for equal lengths.
        :param old_axis_id: axis index (int) or name (str)
        :param new_axis: Series or Index object
        :return: new Axes object
        """
        old_axis_index = self.index(old_axis_id)
        return self._replace(old_axis_index, new_axis)

    def _replace(self, old_axis_index, new_axis):
        new_axes = list(self.axes)
        new_axes[old_axis_index] = new_axis
        return Axes(new_axes)

    def swap(self, axis_id1, axis_id2):
        """Return a new Axes object with two axes swapped.
        :param axis_id1: name or index of the first axis
        :param axis_id2: name or index of the second axis
        :return: new Axes object
        """
        index1 = self.index(axis_id1)
        index2 = self.index(axis_id2)
        new_axes = list(self)
        new_axes[index1], new_axes[index2] = new_axes[index2], new_axes[index1]
        return Axes(new_axes)


def intersect(axes1, axes2):
    """Return the space intersect of the common axes. The order of values on each axis correspond to the order on
    the corresponding axis on axes1. The result can be used for inner join operations.
    :param axes1: Axes object
    :param axes2: Axes object
    :return: Axes
    """
    common_axes = []
    for axis1 in axes1:
        axis2 = axes2.axis_by_name(axis1.name)
        if axis2 is None:
            continue

        if axis1 is axes2:
            axis = axis1
        else:
            # TODO: assume_unique=True ?
            indices = np.in1d(axis1.values, axis2.values)
            axis = axis1[indices]

        common_axes.append(axis)

    return Axes(common_axes)


def make_axes(axes):
    """Creates an Axes object from a collection of axes."""
    if not isinstance(axes, Axes):
        return Axes(axes)
    else:
        return axes
